"""WebTransport Constants and Defaults."""

from __future__ import annotations

import ssl
from enum import IntEnum
from typing import TYPE_CHECKING, Any, TypedDict

from pywebtransport.types import Headers
from pywebtransport.version import __version__

if TYPE_CHECKING:
    from pywebtransport.config import ProxyConfig


__all__ = [
    "ALPN_H3",
    "ALPN_H3_29",
    "ALPN_H3_32",
    "BIDIRECTIONAL_STREAM",
    "CLOSE_WEBTRANSPORT_SESSION_TYPE",
    "ClientConfigDefaults",
    "DEFAULT_ACCESS_LOG",
    "DEFAULT_ALPN_PROTOCOLS",
    "DEFAULT_AUTO_RECONNECT",
    "DEFAULT_BIND_HOST",
    "DEFAULT_BUFFER_SIZE",
    "DEFAULT_CERTFILE",
    "DEFAULT_CLIENT_MAX_CONNECTIONS",
    "DEFAULT_CLIENT_VERIFY_MODE",
    "DEFAULT_CLOSE_TIMEOUT",
    "DEFAULT_CONNECT_TIMEOUT",
    "DEFAULT_CONGESTION_CONTROL_ALGORITHM",
    "DEFAULT_CONNECTION_CLEANUP_INTERVAL",
    "DEFAULT_CONNECTION_IDLE_CHECK_INTERVAL",
    "DEFAULT_CONNECTION_IDLE_TIMEOUT",
    "DEFAULT_CONNECTION_KEEPALIVE_TIMEOUT",
    "DEFAULT_DEBUG",
    "DEFAULT_DEV_PORT",
    "DEFAULT_FLOW_CONTROL_WINDOW_AUTO_SCALE",
    "DEFAULT_FLOW_CONTROL_WINDOW_SIZE",
    "DEFAULT_INITIAL_MAX_DATA",
    "DEFAULT_INITIAL_MAX_STREAMS_BIDI",
    "DEFAULT_INITIAL_MAX_STREAMS_UNI",
    "DEFAULT_KEEP_ALIVE",
    "DEFAULT_KEYFILE",
    "DEFAULT_LOG_FORMAT",
    "DEFAULT_LOG_LEVEL",
    "DEFAULT_MAX_DATAGRAM_SIZE",
    "DEFAULT_MAX_INCOMING_STREAMS",
    "DEFAULT_MAX_PENDING_EVENTS_PER_SESSION",
    "DEFAULT_MAX_RETRIES",
    "DEFAULT_MAX_RETRY_DELAY",
    "DEFAULT_MAX_SESSIONS",
    "DEFAULT_MAX_STREAMS",
    "DEFAULT_MAX_STREAMS_PER_CONNECTION",
    "DEFAULT_MAX_TOTAL_PENDING_EVENTS",
    "DEFAULT_PENDING_EVENT_TTL",
    "DEFAULT_PORT",
    "DEFAULT_PROXY_CONNECT_TIMEOUT",
    "DEFAULT_PUBSUB_SUBSCRIPTION_QUEUE_SIZE",
    "DEFAULT_READ_TIMEOUT",
    "DEFAULT_RETRY_BACKOFF",
    "DEFAULT_RETRY_DELAY",
    "DEFAULT_SECURE_PORT",
    "DEFAULT_SERVER_MAX_CONNECTIONS",
    "DEFAULT_SERVER_VERIFY_MODE",
    "DEFAULT_SESSION_CLEANUP_INTERVAL",
    "DEFAULT_STREAM_CLEANUP_INTERVAL",
    "DEFAULT_STREAM_CREATION_TIMEOUT",
    "DEFAULT_STREAM_FLOW_CONTROL_INCREMENT_BIDI",
    "DEFAULT_STREAM_FLOW_CONTROL_INCREMENT_UNI",
    "DEFAULT_VERSION",
    "DEFAULT_WEBTRANSPORT_PATH",
    "DEFAULT_WRITE_TIMEOUT",
    "DRAIN_WEBTRANSPORT_SESSION_TYPE",
    "DRAFT_VERSION",
    "Defaults",
    "ErrorCodes",
    "H3_FRAME_TYPE_CANCEL_PUSH",
    "H3_FRAME_TYPE_DATA",
    "H3_FRAME_TYPE_GOAWAY",
    "H3_FRAME_TYPE_HEADERS",
    "H3_FRAME_TYPE_MAX_PUSH_ID",
    "H3_FRAME_TYPE_PUSH_PROMISE",
    "H3_FRAME_TYPE_SETTINGS",
    "H3_FRAME_TYPE_WEBTRANSPORT_STREAM",
    "H3_STREAM_TYPE_CONTROL",
    "H3_STREAM_TYPE_PUSH",
    "H3_STREAM_TYPE_QPACK_DECODER",
    "H3_STREAM_TYPE_QPACK_ENCODER",
    "H3_STREAM_TYPE_WEBTRANSPORT",
    "MAX_BUFFER_SIZE",
    "MAX_DATAGRAM_SIZE",
    "MAX_STREAM_ID",
    "ORIGIN_HEADER",
    "RECOMMENDED_BUFFER_SIZES",
    "SECURE_SCHEMES",
    "SEC_WEBTRANSPORT_HTTP3_DRAFT13",
    "SETTINGS_ENABLE_CONNECT_PROTOCOL",
    "SETTINGS_H3_DATAGRAM",
    "SETTINGS_QPACK_BLOCKED_STREAMS",
    "SETTINGS_QPACK_MAX_TABLE_CAPACITY",
    "SETTINGS_WT_INITIAL_MAX_DATA",
    "SETTINGS_WT_INITIAL_MAX_STREAMS_BIDI",
    "SETTINGS_WT_INITIAL_MAX_STREAMS_UNI",
    "SETTINGS_WT_MAX_SESSIONS",
    "ServerConfigDefaults",
    "SUPPORTED_CONGESTION_CONTROL_ALGORITHMS",
    "SUPPORTED_VERSIONS",
    "UNIDIRECTIONAL_STREAM",
    "USER_AGENT_HEADER",
    "WEBTRANSPORT_H3_BIDI_STREAM_TYPE",
    "WEBTRANSPORT_H3_UNI_STREAM_TYPE",
    "WEBTRANSPORT_HEADER",
    "WEBTRANSPORT_MIME_TYPE",
    "WEBTRANSPORT_SCHEMES",
    "WT_DATA_BLOCKED_TYPE",
    "WT_MAX_DATA_TYPE",
    "WT_MAX_STREAMS_BIDI_TYPE",
    "WT_MAX_STREAMS_UNI_TYPE",
    "WT_STREAMS_BLOCKED_BIDI_TYPE",
    "WT_STREAMS_BLOCKED_UNI_TYPE",
]


# General Web Standard Identifiers
ALPN_H3: str = "h3"
ALPN_H3_29: str = "h3-29"
ALPN_H3_32: str = "h3-32"
ORIGIN_HEADER: str = "origin"
SECURE_SCHEMES: tuple[str, str] = ("https", "wss")
USER_AGENT_HEADER: str = "user-agent"
WEBTRANSPORT_MIME_TYPE: str = "application/webtransport"
WEBTRANSPORT_SCHEMES: tuple[str, str] = ("https", "wss")


# Protocol-Defined Constants
BIDIRECTIONAL_STREAM: int = 0x0
CLOSE_WEBTRANSPORT_SESSION_TYPE: int = 0x2843
DRAIN_WEBTRANSPORT_SESSION_TYPE: int = 0x78AE
DRAFT_VERSION: int = 13
H3_FRAME_TYPE_DATA: int = 0x0
H3_FRAME_TYPE_HEADERS: int = 0x1
H3_FRAME_TYPE_CANCEL_PUSH: int = 0x3
H3_FRAME_TYPE_SETTINGS: int = 0x4
H3_FRAME_TYPE_PUSH_PROMISE: int = 0x5
H3_FRAME_TYPE_GOAWAY: int = 0x7
H3_FRAME_TYPE_MAX_PUSH_ID: int = 0xD
H3_FRAME_TYPE_WEBTRANSPORT_STREAM: int = 0x41
H3_STREAM_TYPE_CONTROL: int = 0x0
H3_STREAM_TYPE_PUSH: int = 0x1
H3_STREAM_TYPE_QPACK_ENCODER: int = 0x2
H3_STREAM_TYPE_QPACK_DECODER: int = 0x3
H3_STREAM_TYPE_WEBTRANSPORT: int = 0x54
MAX_DATAGRAM_SIZE: int = 65535
MAX_STREAM_ID: int = 2**62 - 1
SEC_WEBTRANSPORT_HTTP3_DRAFT13: str = "webtransport"
SETTINGS_ENABLE_CONNECT_PROTOCOL: int = 0x8
SETTINGS_H3_DATAGRAM: int = 0x33
SETTINGS_QPACK_BLOCKED_STREAMS: int = 0x7
SETTINGS_QPACK_MAX_TABLE_CAPACITY: int = 0x1
SETTINGS_WT_INITIAL_MAX_DATA: int = 0x2B61
SETTINGS_WT_INITIAL_MAX_STREAMS_BIDI: int = 0x2B65
SETTINGS_WT_INITIAL_MAX_STREAMS_UNI: int = 0x2B64
SETTINGS_WT_MAX_SESSIONS: int = 0x14E9CD29
UNIDIRECTIONAL_STREAM: int = 0x2
WEBTRANSPORT_H3_BIDI_STREAM_TYPE: int = 0x41
WEBTRANSPORT_H3_UNI_STREAM_TYPE: int = 0x54
WEBTRANSPORT_HEADER: str = "webtransport"
WT_DATA_BLOCKED_TYPE: int = 0x190B4D41
WT_MAX_DATA_TYPE: int = 0x190B4D3D
WT_MAX_STREAMS_BIDI_TYPE: int = 0x190B4D3F
WT_MAX_STREAMS_UNI_TYPE: int = 0x190B4D40
WT_STREAMS_BLOCKED_BIDI_TYPE: int = 0x190B4D43
WT_STREAMS_BLOCKED_UNI_TYPE: int = 0x190B4D44


# Library Defaults & Utils
DEFAULT_ACCESS_LOG: bool = True
DEFAULT_ALPN_PROTOCOLS: tuple[str, str] = (ALPN_H3, ALPN_H3_29)
DEFAULT_AUTO_RECONNECT: bool = False
DEFAULT_BIND_HOST: str = "localhost"
DEFAULT_BUFFER_SIZE: int = 65536
DEFAULT_CERTFILE: str = ""
DEFAULT_CLIENT_MAX_CONNECTIONS: int = 100
DEFAULT_CLIENT_VERIFY_MODE: ssl.VerifyMode = ssl.CERT_REQUIRED
DEFAULT_CLOSE_TIMEOUT: float = 5.0
DEFAULT_CONNECT_TIMEOUT: float = 30.0
DEFAULT_CONGESTION_CONTROL_ALGORITHM: str = "cubic"
DEFAULT_CONNECTION_CLEANUP_INTERVAL: float = 30.0
DEFAULT_CONNECTION_IDLE_CHECK_INTERVAL: float = 5.0
DEFAULT_CONNECTION_IDLE_TIMEOUT: float = 60.0
DEFAULT_CONNECTION_KEEPALIVE_TIMEOUT: float = 30.0
DEFAULT_DEBUG: bool = False
DEFAULT_DEV_PORT: int = 4433
DEFAULT_FLOW_CONTROL_WINDOW_AUTO_SCALE: bool = True
DEFAULT_FLOW_CONTROL_WINDOW_SIZE: int = 1024 * 1024
DEFAULT_INITIAL_MAX_DATA: int = 0
DEFAULT_INITIAL_MAX_STREAMS_BIDI: int = 0
DEFAULT_INITIAL_MAX_STREAMS_UNI: int = 0
DEFAULT_KEEP_ALIVE: bool = True
DEFAULT_KEYFILE: str = ""
DEFAULT_LOG_FORMAT: str = "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
DEFAULT_LOG_LEVEL: str = "INFO"
DEFAULT_MAX_DATAGRAM_SIZE: int = 65535
DEFAULT_MAX_INCOMING_STREAMS: int = 100
DEFAULT_MAX_PENDING_EVENTS_PER_SESSION: int = 16
DEFAULT_MAX_RETRIES: int = 3
DEFAULT_MAX_RETRY_DELAY: float = 30.0
DEFAULT_MAX_SESSIONS: int = 10000
DEFAULT_MAX_STREAMS: int = 100
DEFAULT_MAX_STREAMS_PER_CONNECTION: int = 100
DEFAULT_MAX_TOTAL_PENDING_EVENTS: int = 1000
DEFAULT_PENDING_EVENT_TTL: float = 5.0
DEFAULT_PORT: int = 80
DEFAULT_PROXY_CONNECT_TIMEOUT: float = 10.0
DEFAULT_PUBSUB_SUBSCRIPTION_QUEUE_SIZE: int = 16
DEFAULT_READ_TIMEOUT: float = 60.0
DEFAULT_RETRY_BACKOFF: float = 2.0
DEFAULT_RETRY_DELAY: float = 1.0
DEFAULT_SECURE_PORT: int = 443
DEFAULT_SERVER_MAX_CONNECTIONS: int = 3000
DEFAULT_SERVER_VERIFY_MODE: ssl.VerifyMode = ssl.CERT_NONE
DEFAULT_SESSION_CLEANUP_INTERVAL: float = 60.0
DEFAULT_STREAM_CLEANUP_INTERVAL: float = 15.0
DEFAULT_STREAM_CREATION_TIMEOUT: float = 10.0
DEFAULT_STREAM_FLOW_CONTROL_INCREMENT_BIDI: int = 10
DEFAULT_STREAM_FLOW_CONTROL_INCREMENT_UNI: int = 10
DEFAULT_VERSION: str = "h3"
DEFAULT_WEBTRANSPORT_PATH: str = "/webtransport"
DEFAULT_WRITE_TIMEOUT: float = 30.0
MAX_BUFFER_SIZE: int = 1024 * 1024
RECOMMENDED_BUFFER_SIZES: dict[str, int] = {"small": 8192, "medium": 65536, "large": 262144}
SUPPORTED_CONGESTION_CONTROL_ALGORITHMS: tuple[str, str] = ("reno", "cubic")
SUPPORTED_VERSIONS: tuple[str, ...] = ("draft-ietf-webtrans-http3-13", "h3")


class ErrorCodes(IntEnum):
    """A collection of standard WebTransport and QUIC error codes."""

    NO_ERROR = 0x0
    INTERNAL_ERROR = 0x1
    CONNECTION_REFUSED = 0x2
    FLOW_CONTROL_ERROR = 0x3
    STREAM_LIMIT_ERROR = 0x4
    STREAM_STATE_ERROR = 0x5
    FINAL_SIZE_ERROR = 0x6
    FRAME_ENCODING_ERROR = 0x7
    TRANSPORT_PARAMETER_ERROR = 0x8
    CONNECTION_ID_LIMIT_ERROR = 0x9
    PROTOCOL_VIOLATION = 0xA
    INVALID_TOKEN = 0xB
    APPLICATION_ERROR = 0xC
    CRYPTO_BUFFER_EXCEEDED = 0xD
    KEY_UPDATE_ERROR = 0xE
    AEAD_LIMIT_REACHED = 0xF
    NO_VIABLE_PATH = 0x10
    H3_DATAGRAM_ERROR = 0x33
    H3_NO_ERROR = 0x100
    H3_GENERAL_PROTOCOL_ERROR = 0x101
    H3_INTERNAL_ERROR = 0x102
    H3_STREAM_CREATION_ERROR = 0x103
    H3_CLOSED_CRITICAL_STREAM = 0x104
    H3_FRAME_UNEXPECTED = 0x105
    H3_FRAME_ERROR = 0x106
    H3_EXCESSIVE_LOAD = 0x107
    H3_ID_ERROR = 0x108
    H3_SETTINGS_ERROR = 0x109
    H3_MISSING_SETTINGS = 0x10A
    H3_REQUEST_REJECTED = 0x10B
    H3_REQUEST_CANCELLED = 0x10C
    H3_REQUEST_INCOMPLETE = 0x10D
    H3_MESSAGE_ERROR = 0x10E
    H3_CONNECT_ERROR = 0x10F
    H3_VERSION_FALLBACK = 0x110
    WT_SESSION_GONE = 0x170D7B68
    WT_BUFFERED_STREAM_REJECTED = 0x3994BD84
    WT_APPLICATION_ERROR_FIRST = 0x52E4A40FA8DB
    QPACK_DECOMPRESSION_FAILED = 0x200
    QPACK_ENCODER_STREAM_ERROR = 0x201
    QPACK_DECODER_STREAM_ERROR = 0x202
    APP_CONNECTION_TIMEOUT = 0x1000
    APP_AUTHENTICATION_FAILED = 0x1001
    APP_PERMISSION_DENIED = 0x1002
    APP_RESOURCE_EXHAUSTED = 0x1003
    APP_INVALID_REQUEST = 0x1004
    APP_SERVICE_UNAVAILABLE = 0x1005


class ClientConfigDefaults(TypedDict):
    """A type definition for the client configuration dictionary."""

    alpn_protocols: list[str]
    auto_reconnect: bool
    ca_certs: str | None
    certfile: str | None
    close_timeout: float
    congestion_control_algorithm: str
    connect_timeout: float
    connection_cleanup_interval: float
    connection_idle_check_interval: float
    connection_idle_timeout: float
    connection_keepalive_timeout: float
    debug: bool
    flow_control_window_auto_scale: bool
    flow_control_window_size: int
    headers: Headers
    initial_max_data: int
    initial_max_streams_bidi: int
    initial_max_streams_uni: int
    keep_alive: bool
    keyfile: str | None
    log_level: str
    max_connections: int
    max_datagram_size: int
    max_incoming_streams: int
    max_pending_events_per_session: int
    max_retries: int
    max_retry_delay: float
    max_stream_buffer_size: int
    max_streams: int
    max_total_pending_events: int
    pending_event_ttl: float
    proxy: ProxyConfig | None
    read_timeout: float | None
    retry_backoff: float
    retry_delay: float
    stream_buffer_size: int
    stream_cleanup_interval: float
    stream_creation_timeout: float
    stream_flow_control_increment_bidi: int
    stream_flow_control_increment_uni: int
    user_agent: str
    verify_mode: ssl.VerifyMode | None
    write_timeout: float | None


class ServerConfigDefaults(TypedDict):
    """A type definition for the server configuration dictionary."""

    access_log: bool
    alpn_protocols: list[str]
    bind_host: str
    bind_port: int
    ca_certs: str | None
    certfile: str
    congestion_control_algorithm: str
    connection_cleanup_interval: float
    connection_idle_check_interval: float
    connection_idle_timeout: float
    connection_keepalive_timeout: float
    debug: bool
    flow_control_window_auto_scale: bool
    flow_control_window_size: int
    initial_max_data: int
    initial_max_streams_bidi: int
    initial_max_streams_uni: int
    keep_alive: bool
    keyfile: str
    log_level: str
    max_connections: int
    max_datagram_size: int
    max_incoming_streams: int
    max_pending_events_per_session: int
    max_sessions: int
    max_stream_buffer_size: int
    max_streams_per_connection: int
    max_total_pending_events: int
    middleware: list[Any]
    pending_event_ttl: float
    read_timeout: float | None
    session_cleanup_interval: float
    stream_buffer_size: int
    stream_cleanup_interval: float
    stream_flow_control_increment_bidi: int
    stream_flow_control_increment_uni: int
    verify_mode: ssl.VerifyMode
    write_timeout: float | None


_DEFAULT_CLIENT_CONFIG: ClientConfigDefaults = {
    "alpn_protocols": list(DEFAULT_ALPN_PROTOCOLS),
    "auto_reconnect": DEFAULT_AUTO_RECONNECT,
    "ca_certs": None,
    "certfile": None,
    "close_timeout": DEFAULT_CLOSE_TIMEOUT,
    "congestion_control_algorithm": DEFAULT_CONGESTION_CONTROL_ALGORITHM,
    "connect_timeout": DEFAULT_CONNECT_TIMEOUT,
    "connection_cleanup_interval": DEFAULT_CONNECTION_CLEANUP_INTERVAL,
    "connection_idle_check_interval": DEFAULT_CONNECTION_IDLE_CHECK_INTERVAL,
    "connection_idle_timeout": DEFAULT_CONNECTION_IDLE_TIMEOUT,
    "connection_keepalive_timeout": DEFAULT_CONNECTION_KEEPALIVE_TIMEOUT,
    "debug": DEFAULT_DEBUG,
    "flow_control_window_auto_scale": DEFAULT_FLOW_CONTROL_WINDOW_AUTO_SCALE,
    "flow_control_window_size": DEFAULT_FLOW_CONTROL_WINDOW_SIZE,
    "headers": {},
    "initial_max_data": DEFAULT_INITIAL_MAX_DATA,
    "initial_max_streams_bidi": DEFAULT_INITIAL_MAX_STREAMS_BIDI,
    "initial_max_streams_uni": DEFAULT_INITIAL_MAX_STREAMS_UNI,
    "keep_alive": DEFAULT_KEEP_ALIVE,
    "keyfile": None,
    "log_level": DEFAULT_LOG_LEVEL,
    "max_connections": DEFAULT_CLIENT_MAX_CONNECTIONS,
    "max_datagram_size": DEFAULT_MAX_DATAGRAM_SIZE,
    "max_incoming_streams": DEFAULT_MAX_INCOMING_STREAMS,
    "max_pending_events_per_session": DEFAULT_MAX_PENDING_EVENTS_PER_SESSION,
    "max_retries": DEFAULT_MAX_RETRIES,
    "max_retry_delay": DEFAULT_MAX_RETRY_DELAY,
    "max_stream_buffer_size": MAX_BUFFER_SIZE,
    "max_streams": DEFAULT_MAX_STREAMS,
    "max_total_pending_events": DEFAULT_MAX_TOTAL_PENDING_EVENTS,
    "pending_event_ttl": DEFAULT_PENDING_EVENT_TTL,
    "proxy": None,
    "read_timeout": DEFAULT_READ_TIMEOUT,
    "retry_backoff": DEFAULT_RETRY_BACKOFF,
    "retry_delay": DEFAULT_RETRY_DELAY,
    "stream_buffer_size": DEFAULT_BUFFER_SIZE,
    "stream_cleanup_interval": DEFAULT_STREAM_CLEANUP_INTERVAL,
    "stream_creation_timeout": DEFAULT_STREAM_CREATION_TIMEOUT,
    "stream_flow_control_increment_bidi": DEFAULT_STREAM_FLOW_CONTROL_INCREMENT_BIDI,
    "stream_flow_control_increment_uni": DEFAULT_STREAM_FLOW_CONTROL_INCREMENT_UNI,
    "user_agent": f"pywebtransport/{__version__}",
    "verify_mode": DEFAULT_CLIENT_VERIFY_MODE,
    "write_timeout": DEFAULT_WRITE_TIMEOUT,
}

_DEFAULT_SERVER_CONFIG: ServerConfigDefaults = {
    "access_log": DEFAULT_ACCESS_LOG,
    "alpn_protocols": list(DEFAULT_ALPN_PROTOCOLS),
    "bind_host": DEFAULT_BIND_HOST,
    "bind_port": DEFAULT_DEV_PORT,
    "ca_certs": None,
    "certfile": DEFAULT_CERTFILE,
    "congestion_control_algorithm": DEFAULT_CONGESTION_CONTROL_ALGORITHM,
    "connection_cleanup_interval": DEFAULT_CONNECTION_CLEANUP_INTERVAL,
    "connection_idle_check_interval": DEFAULT_CONNECTION_IDLE_CHECK_INTERVAL,
    "connection_idle_timeout": DEFAULT_CONNECTION_IDLE_TIMEOUT,
    "connection_keepalive_timeout": DEFAULT_CONNECTION_KEEPALIVE_TIMEOUT,
    "debug": DEFAULT_DEBUG,
    "flow_control_window_auto_scale": DEFAULT_FLOW_CONTROL_WINDOW_AUTO_SCALE,
    "flow_control_window_size": DEFAULT_FLOW_CONTROL_WINDOW_SIZE,
    "initial_max_data": DEFAULT_INITIAL_MAX_DATA,
    "initial_max_streams_bidi": DEFAULT_INITIAL_MAX_STREAMS_BIDI,
    "initial_max_streams_uni": DEFAULT_INITIAL_MAX_STREAMS_UNI,
    "keep_alive": DEFAULT_KEEP_ALIVE,
    "keyfile": DEFAULT_KEYFILE,
    "log_level": DEFAULT_LOG_LEVEL,
    "max_connections": DEFAULT_SERVER_MAX_CONNECTIONS,
    "max_datagram_size": DEFAULT_MAX_DATAGRAM_SIZE,
    "max_incoming_streams": DEFAULT_MAX_INCOMING_STREAMS,
    "max_pending_events_per_session": DEFAULT_MAX_PENDING_EVENTS_PER_SESSION,
    "max_sessions": DEFAULT_MAX_SESSIONS,
    "max_stream_buffer_size": MAX_BUFFER_SIZE,
    "max_streams_per_connection": DEFAULT_MAX_STREAMS_PER_CONNECTION,
    "max_total_pending_events": DEFAULT_MAX_TOTAL_PENDING_EVENTS,
    "middleware": [],
    "pending_event_ttl": DEFAULT_PENDING_EVENT_TTL,
    "read_timeout": DEFAULT_READ_TIMEOUT,
    "session_cleanup_interval": DEFAULT_SESSION_CLEANUP_INTERVAL,
    "stream_buffer_size": DEFAULT_BUFFER_SIZE,
    "stream_cleanup_interval": DEFAULT_STREAM_CLEANUP_INTERVAL,
    "stream_flow_control_increment_bidi": DEFAULT_STREAM_FLOW_CONTROL_INCREMENT_BIDI,
    "stream_flow_control_increment_uni": DEFAULT_STREAM_FLOW_CONTROL_INCREMENT_UNI,
    "verify_mode": DEFAULT_SERVER_VERIFY_MODE,
    "write_timeout": DEFAULT_WRITE_TIMEOUT,
}


class Defaults:
    """Provides access to default configuration values."""

    @staticmethod
    def get_client_config() -> ClientConfigDefaults:
        """Return a copy of the default client configuration."""
        return _DEFAULT_CLIENT_CONFIG.copy()

    @staticmethod
    def get_server_config() -> ServerConfigDefaults:
        """Return a copy of the default server configuration."""
        return _DEFAULT_SERVER_CONFIG.copy()



"""WebTransport Request Router."""

from __future__ import annotations

import re
from typing import Any, Pattern

from pywebtransport.session import WebTransportSession
from pywebtransport.types import SessionHandler
from pywebtransport.utils import get_logger

__all__ = ["RequestRouter"]

logger = get_logger(name="server.router")


class RequestRouter:
    """Routes session requests to handlers based on path matching."""

    def __init__(self) -> None:
        """Initialize the request router."""
        self._routes: dict[str, SessionHandler] = {}
        self._pattern_routes: list[tuple[Pattern[str], SessionHandler]] = []
        self._default_handler: SessionHandler | None = None

    def add_pattern_route(self, *, pattern: str, handler: SessionHandler) -> None:
        """Add a route for a regular expression pattern."""
        try:
            compiled_pattern = re.compile(pattern)
            self._pattern_routes.append((compiled_pattern, handler))
            logger.debug("Added pattern route: %s", pattern)
        except re.error as e:
            logger.error("Invalid regex pattern '%s': %s", pattern, e, exc_info=True)
            raise

    def add_route(self, *, path: str, handler: SessionHandler) -> None:
        """Add a route for an exact path match."""
        self._routes[path] = handler
        logger.debug("Added route: %s", path)

    def remove_route(self, *, path: str) -> None:
        """Remove a route for an exact path match."""
        if path in self._routes:
            del self._routes[path]
            logger.debug("Removed route: %s", path)

    def route_request(self, *, session: WebTransportSession) -> SessionHandler | None:
        """Route a request to the appropriate handler based on the session's path."""
        path = session.path
        handler: SessionHandler | None = None

        if path in self._routes:
            handler = self._routes[path]
        else:
            for pattern, pattern_handler in self._pattern_routes:
                match = pattern.match(path)
                if match:
                    setattr(session, "path_params", match.groups())
                    handler = pattern_handler
                    break

        return handler or self._default_handler

    def set_default_handler(self, *, handler: SessionHandler) -> None:
        """Set a default handler for routes that are not matched."""
        self._default_handler = handler
        logger.debug("Set default handler")

    def get_all_routes(self) -> dict[str, SessionHandler]:
        """Get a copy of all registered exact-match routes."""
        return self._routes.copy()

    def get_route_handler(self, *, path: str) -> SessionHandler | None:
        """Get the handler for a specific path (exact match only)."""
        return self._routes.get(path)

    def get_route_stats(self) -> dict[str, Any]:
        """Get statistics about the configured routes."""
        return {
            "exact_routes": len(self._routes),
            "pattern_routes": len(self._pattern_routes),
            "has_default_handler": self._default_handler is not None,
        }



"""WebTransport Session Subpackage."""

from .manager import SessionManager
from .session import SessionStats, WebTransportSession

__all__ = [
    "SessionManager",
    "SessionStats",
    "WebTransportSession",
]



"""WebTransport Server Monitor."""

from __future__ import annotations

import asyncio
from collections import deque
from types import TracebackType
from typing import Any, Self

from pywebtransport.server.server import WebTransportServer
from pywebtransport.utils import get_logger, get_timestamp

__all__ = ["ServerMonitor"]

logger = get_logger(name="server.monitor")


class ServerMonitor:
    """Monitors server performance and health via an async context."""

    def __init__(self, server: WebTransportServer, *, monitoring_interval: float = 30.0) -> None:
        """Initialize the server monitor."""
        self._server = server
        self._interval = monitoring_interval
        self._monitor_task: asyncio.Task[None] | None = None
        self._metrics_history: deque[dict[str, Any]] = deque(maxlen=120)
        self._alerts: deque[dict[str, Any]] = deque(maxlen=100)

    @classmethod
    def create(cls, *, server: WebTransportServer, monitoring_interval: float = 30.0) -> Self:
        """Factory method to create a new server monitor instance."""
        return cls(server=server, monitoring_interval=monitoring_interval)

    @property
    def is_monitoring(self) -> bool:
        """Check if the monitoring task is currently active."""
        return self._monitor_task is not None and not self._monitor_task.done()

    async def __aenter__(self) -> Self:
        """Enter the async context and start the monitoring task."""
        if self.is_monitoring:
            return self
        try:
            self._monitor_task = asyncio.create_task(self._monitor_loop())
            logger.info("Server monitoring started.")
        except RuntimeError:
            logger.error(
                "Failed to start server monitor: No running event loop.",
                exc_info=True,
            )
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        """Exit the async context and stop the monitoring task."""
        if self._monitor_task:
            self._monitor_task.cancel()
            try:
                await self._monitor_task
            except asyncio.CancelledError:
                pass
        logger.info("Server monitoring stopped.")

    def get_health_status(self) -> dict[str, Any]:
        """Get the current server health status based on the latest metrics."""
        metrics = self.get_current_metrics()
        if not metrics:
            return {"status": "unknown", "reason": "No metrics collected yet."}

        stats = metrics.get("stats", {})
        if not self._server.is_serving:
            return {"status": "unhealthy", "reason": "Server is not serving."}

        connections = stats.get("connections", {})
        accepted = stats.get("connections_accepted", 0)
        rejected = stats.get("connections_rejected", 0)
        total_attempts = accepted + rejected

        if total_attempts > 10:
            success_rate = accepted / total_attempts if total_attempts > 0 else 1.0
            if success_rate < 0.9:
                return {
                    "status": "degraded",
                    "reason": f"Low connection success rate: {success_rate:.2%}",
                }

        if connections and connections.get("active", 0) > 0:
            return {"status": "healthy", "reason": "Server is operating normally."}

        return {"status": "idle", "reason": "Server is running but has no active connections."}

    def get_alerts(self, *, limit: int = 25) -> list[dict[str, Any]]:
        """Get a list of recently generated alerts."""
        return list(self._alerts)[-limit:]

    def get_current_metrics(self) -> dict[str, Any] | None:
        """Get the latest collected metrics."""
        return self._metrics_history[-1] if self._metrics_history else None

    def get_metrics_history(self, *, limit: int = 100) -> list[dict[str, Any]]:
        """Get a list of recent metrics history."""
        return list(self._metrics_history)[-limit:]

    def clear_history(self) -> None:
        """Clear all collected metrics and alerts history."""
        self._metrics_history.clear()
        self._alerts.clear()
        logger.info("Metrics and alerts history cleared.")

    async def _monitor_loop(self) -> None:
        """Run the main loop for periodically collecting metrics and checking health."""
        try:
            while True:
                await self._collect_metrics()
                await self._check_for_alerts()
                await asyncio.sleep(self._interval)
        except asyncio.CancelledError:
            logger.info("Server monitor loop has been cancelled.")
        except Exception as e:
            logger.error("Server monitor loop encountered a critical error: %s", e, exc_info=True)

    async def _collect_metrics(self) -> None:
        """Collect a snapshot of the server's current statistics."""
        try:
            timestamp = get_timestamp()
            stats = await self._server.get_server_stats()
            metrics = {"timestamp": timestamp, "stats": stats}
            self._metrics_history.append(metrics)
        except Exception as e:
            logger.error("Metrics collection failed: %s", e, exc_info=True)

    async def _check_for_alerts(self) -> None:
        """Analyze the latest metrics and generate alerts if thresholds are breached."""
        try:
            health = self.get_health_status()
            if health["status"] in ("unhealthy", "degraded"):
                if not self._alerts or self._alerts[-1].get("reason") != health["reason"]:
                    alert = {
                        "timestamp": get_timestamp(),
                        "status": health["status"],
                        "reason": health["reason"],
                    }
                    self._alerts.append(alert)
                    logger.warning("Health Alert: %s - %s", health["status"], health["reason"])
        except Exception as e:
            logger.error("Alert check failed: %s", e, exc_info=True)



"""WebTransport Client Utilities."""

from __future__ import annotations

import asyncio
from typing import Any

from pywebtransport.client.client import WebTransportClient
from pywebtransport.config import ClientConfig
from pywebtransport.utils import get_logger, get_timestamp

__all__ = [
    "benchmark_client_performance",
    "test_client_connectivity",
]

logger = get_logger(name="client.utils")


async def benchmark_client_performance(
    *,
    url: str,
    config: ClientConfig | None = None,
    num_requests: int = 100,
    concurrent_requests: int = 10,
) -> dict[str, Any]:
    """Benchmark client performance by measuring request-response latency."""
    results: dict[str, Any] = {
        "total_requests": num_requests,
        "successful_requests": 0,
        "failed_requests": 0,
        "latencies": [],
    }
    semaphore = asyncio.Semaphore(value=concurrent_requests)

    async def benchmark_single_request(*, client: WebTransportClient) -> float:
        async with semaphore:
            start_time = get_timestamp()
            session = await client.connect(url=url)
            try:
                stream = await session.create_bidirectional_stream()
                await stream.write(data=b"benchmark_ping")
                _ = await stream.read(size=1024)
                latency = get_timestamp() - start_time
                return latency
            finally:
                if session and not session.is_closed:
                    await session.close()

    tasks: list[asyncio.Task[float]] = []
    async with WebTransportClient(config=config) as client:
        try:
            async with asyncio.TaskGroup() as tg:
                for _ in range(num_requests):
                    tasks.append(tg.create_task(benchmark_single_request(client=client)))
        except* Exception as eg:
            logger.warning("%d benchmark requests failed.", len(eg.exceptions), exc_info=eg)

    for task in tasks:
        if task.done() and not task.exception():
            results["successful_requests"] += 1
            results["latencies"].append(task.result())
        else:
            results["failed_requests"] += 1

    if results["latencies"]:
        latencies = results["latencies"]
        results["avg_latency"] = sum(latencies) / len(latencies)
        results["min_latency"] = min(latencies)
        results["max_latency"] = max(latencies)

    return results


async def test_client_connectivity(
    *, url: str, config: ClientConfig | None = None, timeout: float = 10.0
) -> dict[str, Any]:
    """Test client connectivity to a given URL."""
    async with WebTransportClient(config=config) as client:
        try:
            start_time = get_timestamp()
            session = await asyncio.wait_for(client.connect(url=url, timeout=timeout), timeout=timeout)
            connect_time = get_timestamp() - start_time
            await session.close()
            return {
                "success": True,
                "connect_time": connect_time,
                "client_stats": client.stats,
            }
        except Exception as e:
            return {
                "success": False,
                "error": str(e),
                "connect_time": None,
                "client_stats": client.stats,
            }



"""WebTransport Datagram Broadcaster."""

from __future__ import annotations

import asyncio
from types import TracebackType
from typing import TYPE_CHECKING, Self

from pywebtransport.exceptions import DatagramError
from pywebtransport.types import Data
from pywebtransport.utils import get_logger

if TYPE_CHECKING:
    from pywebtransport.datagram.transport import WebTransportDatagramTransport


__all__ = ["DatagramBroadcaster"]

logger = get_logger(name="datagram.broadcaster")


class DatagramBroadcaster:
    """A broadcaster to send datagrams to multiple transports concurrently."""

    def __init__(self) -> None:
        """Initialize the datagram broadcaster."""
        self._transports: list[WebTransportDatagramTransport] = []
        self._lock: asyncio.Lock | None = None

    @classmethod
    def create(cls) -> Self:
        """Factory method to create a new datagram broadcaster instance."""
        return cls()

    async def __aenter__(self) -> Self:
        """Enter async context, initializing asyncio resources."""
        self._lock = asyncio.Lock()
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        """Exit async context, clearing the transport list."""
        if self._lock:
            async with self._lock:
                self._transports.clear()

    async def add_transport(self, *, transport: WebTransportDatagramTransport) -> None:
        """Add a transport to the broadcast list."""
        if self._lock is None:
            raise DatagramError(
                message=(
                    "DatagramBroadcaster has not been activated. It must be used as an "
                    "asynchronous context manager (`async with ...`)."
                )
            )
        async with self._lock:
            if transport not in self._transports:
                self._transports.append(transport)

    async def broadcast(self, *, data: Data, priority: int = 0, ttl: float | None = None) -> int:
        """Broadcast a datagram to all registered transports concurrently."""
        if self._lock is None:
            raise DatagramError(
                message=(
                    "DatagramBroadcaster has not been activated. It must be used as an "
                    "asynchronous context manager (`async with ...`)."
                )
            )

        sent_count = 0
        failed_transports = []

        async with self._lock:
            transports_copy = self._transports.copy()

        active_transports = []
        tasks = []
        for transport in transports_copy:
            if not transport.is_closed:
                tasks.append(transport.send(data=data, priority=priority, ttl=ttl))
                active_transports.append(transport)
            else:
                failed_transports.append(transport)

        if tasks:
            results = await asyncio.gather(*tasks, return_exceptions=True)
            for transport, result in zip(active_transports, results):
                if isinstance(result, Exception):
                    logger.warning(
                        "Failed to broadcast to transport %s: %s",
                        transport,
                        result,
                        exc_info=True,
                    )
                    failed_transports.append(transport)
                else:
                    sent_count += 1

        if failed_transports:
            async with self._lock:
                for transport in failed_transports:
                    if transport in self._transports:
                        self._transports.remove(transport)

        return sent_count

    async def remove_transport(self, *, transport: WebTransportDatagramTransport) -> None:
        """Remove a transport from the broadcast list."""
        if self._lock is None:
            raise DatagramError(
                message=(
                    "DatagramBroadcaster has not been activated. It must be used as an "
                    "asynchronous context manager (`async with ...`)."
                )
            )
        async with self._lock:
            try:
                self._transports.remove(transport)
            except ValueError:
                pass

    async def get_transport_count(self) -> int:
        """Get the current number of active transports safely."""
        if self._lock is None:
            raise DatagramError(
                message=(
                    "DatagramBroadcaster has not been activated. It must be used as an "
                    "asynchronous context manager (`async with ...`)."
                )
            )
        async with self._lock:
            return len(self._transports)



"""WebTransport Stream Subpackage."""

from .manager import StreamManager
from .pool import StreamPool
from .stream import StreamBuffer, StreamStats, WebTransportReceiveStream, WebTransportSendStream, WebTransportStream
from .structured import StructuredStream

__all__ = [
    "StreamBuffer",
    "StreamManager",
    "StreamPool",
    "StreamStats",
    "StructuredStream",
    "WebTransportReceiveStream",
    "WebTransportSendStream",
    "WebTransportStream",
]



"""High-level structured data stream for WebTransport."""

from __future__ import annotations

import asyncio
import struct
from typing import TYPE_CHECKING, Any

from pywebtransport.exceptions import SerializationError, StreamError
from pywebtransport.types import Serializer

if TYPE_CHECKING:
    from pywebtransport.stream.stream import WebTransportStream


__all__ = ["StructuredStream"]


class StructuredStream:
    """A high-level wrapper for sending and receiving structured objects."""

    _HEADER_FORMAT = "!HI"
    _HEADER_SIZE = struct.calcsize(_HEADER_FORMAT)

    def __init__(
        self,
        *,
        stream: WebTransportStream,
        serializer: Serializer,
        registry: dict[int, type[Any]],
    ) -> None:
        """Initialize the structured stream."""
        self._stream = stream
        self._serializer = serializer
        self._registry = registry
        self._class_to_id = {v: k for k, v in registry.items()}

    @property
    def is_closed(self) -> bool:
        """Check if the underlying stream is closed."""
        return self._stream.is_closed

    @property
    def stream_id(self) -> int:
        """Get the underlying stream's ID."""
        return self._stream.stream_id

    async def close(self) -> None:
        """Close the underlying stream."""
        await self._stream.close()

    async def receive_obj(self) -> Any:
        """Receive, deserialize, and return a Python object from the stream."""
        try:
            header_bytes = await self._stream.readexactly(n=self._HEADER_SIZE)
        except asyncio.IncompleteReadError as e:
            raise StreamError(message="Stream closed while waiting for message header.") from e

        type_id, payload_len = struct.unpack(self._HEADER_FORMAT, header_bytes)
        message_class = self._registry.get(type_id)
        if message_class is None:
            raise SerializationError(message=f"Received unknown message type ID: {type_id}")

        try:
            payload = await self._stream.readexactly(n=payload_len)
        except asyncio.IncompleteReadError as e:
            raise StreamError(
                message=(
                    f"Stream closed prematurely while reading payload of size {payload_len} " f"for type ID {type_id}."
                )
            ) from e

        return self._serializer.deserialize(data=payload, obj_type=message_class)

    async def send_obj(self, *, obj: Any) -> None:
        """Serialize and send a Python object over the stream."""
        obj_type = type(obj)
        type_id = self._class_to_id.get(obj_type)
        if type_id is None:
            raise SerializationError(message=f"Object of type '{obj_type.__name__}' is not registered.")

        payload = self._serializer.serialize(obj=obj)
        payload_len = len(payload)
        header = struct.pack(self._HEADER_FORMAT, type_id, payload_len)

        await self._stream.write(data=header + payload)

    def __aiter__(self) -> "StructuredStream":
        """Return self as the asynchronous iterator."""
        return self

    async def __anext__(self) -> Any:
        """Receive the next object in the async iteration."""
        try:
            return await self.receive_obj()
        except StreamError:
            raise StopAsyncIteration



"""WebTransport Server Subpackage."""

from .app import ServerApp
from .cluster import ServerCluster
from .middleware import (
    MiddlewareManager,
    create_auth_middleware,
    create_cors_middleware,
    create_logging_middleware,
    create_rate_limit_middleware,
)
from .monitor import ServerMonitor
from .router import RequestRouter
from .server import ServerStats, WebTransportServer
from .utils import (
    create_development_server,
    create_echo_server_app,
    create_simple_app,
    echo_handler,
    health_check_handler,
)

__all__ = [
    "MiddlewareManager",
    "RequestRouter",
    "ServerApp",
    "ServerCluster",
    "ServerMonitor",
    "ServerStats",
    "WebTransportServer",
    "create_auth_middleware",
    "create_cors_middleware",
    "create_development_server",
    "create_echo_server_app",
    "create_logging_middleware",
    "create_rate_limit_middleware",
    "create_simple_app",
    "echo_handler",
    "health_check_handler",
]



"""WebTransport Configuration."""

from __future__ import annotations

import copy
import ssl
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Self

from pywebtransport.constants import (
    DEFAULT_ACCESS_LOG,
    DEFAULT_ALPN_PROTOCOLS,
    DEFAULT_AUTO_RECONNECT,
    DEFAULT_BIND_HOST,
    DEFAULT_BUFFER_SIZE,
    DEFAULT_CERTFILE,
    DEFAULT_CLIENT_MAX_CONNECTIONS,
    DEFAULT_CLIENT_VERIFY_MODE,
    DEFAULT_CLOSE_TIMEOUT,
    DEFAULT_CONGESTION_CONTROL_ALGORITHM,
    DEFAULT_CONNECT_TIMEOUT,
    DEFAULT_CONNECTION_CLEANUP_INTERVAL,
    DEFAULT_CONNECTION_IDLE_CHECK_INTERVAL,
    DEFAULT_CONNECTION_IDLE_TIMEOUT,
    DEFAULT_CONNECTION_KEEPALIVE_TIMEOUT,
    DEFAULT_DEBUG,
    DEFAULT_DEV_PORT,
    DEFAULT_FLOW_CONTROL_WINDOW_AUTO_SCALE,
    DEFAULT_FLOW_CONTROL_WINDOW_SIZE,
    DEFAULT_INITIAL_MAX_DATA,
    DEFAULT_INITIAL_MAX_STREAMS_BIDI,
    DEFAULT_INITIAL_MAX_STREAMS_UNI,
    DEFAULT_KEEP_ALIVE,
    DEFAULT_KEYFILE,
    DEFAULT_LOG_LEVEL,
    DEFAULT_MAX_DATAGRAM_SIZE,
    DEFAULT_MAX_INCOMING_STREAMS,
    DEFAULT_MAX_PENDING_EVENTS_PER_SESSION,
    DEFAULT_MAX_RETRIES,
    DEFAULT_MAX_RETRY_DELAY,
    DEFAULT_MAX_SESSIONS,
    DEFAULT_MAX_STREAMS,
    DEFAULT_MAX_STREAMS_PER_CONNECTION,
    DEFAULT_MAX_TOTAL_PENDING_EVENTS,
    DEFAULT_PENDING_EVENT_TTL,
    DEFAULT_PROXY_CONNECT_TIMEOUT,
    DEFAULT_READ_TIMEOUT,
    DEFAULT_RETRY_BACKOFF,
    DEFAULT_RETRY_DELAY,
    DEFAULT_SERVER_MAX_CONNECTIONS,
    DEFAULT_SERVER_VERIFY_MODE,
    DEFAULT_SESSION_CLEANUP_INTERVAL,
    DEFAULT_STREAM_CLEANUP_INTERVAL,
    DEFAULT_STREAM_CREATION_TIMEOUT,
    DEFAULT_STREAM_FLOW_CONTROL_INCREMENT_BIDI,
    DEFAULT_STREAM_FLOW_CONTROL_INCREMENT_UNI,
    DEFAULT_WRITE_TIMEOUT,
    MAX_BUFFER_SIZE,
    SUPPORTED_CONGESTION_CONTROL_ALGORITHMS,
    Defaults,
)
from pywebtransport.exceptions import ConfigurationError, certificate_not_found, invalid_config
from pywebtransport.types import Address, Headers, MiddlewareProtocol
from pywebtransport.utils import normalize_headers
from pywebtransport.version import __version__

__all__ = [
    "ClientConfig",
    "ConfigBuilder",
    "ProxyConfig",
    "ServerConfig",
]


@dataclass(kw_only=True)
class ProxyConfig:
    """Configuration for connecting through an HTTP proxy."""

    url: str
    headers: Headers = field(default_factory=dict)
    connect_timeout: float = DEFAULT_PROXY_CONNECT_TIMEOUT


@dataclass(kw_only=True)
class ClientConfig:
    """A comprehensive configuration for the WebTransport client."""

    alpn_protocols: list[str] = field(default_factory=lambda: list(DEFAULT_ALPN_PROTOCOLS))
    auto_reconnect: bool = DEFAULT_AUTO_RECONNECT
    ca_certs: str | None = None
    certfile: str | None = None
    close_timeout: float = DEFAULT_CLOSE_TIMEOUT
    congestion_control_algorithm: str = DEFAULT_CONGESTION_CONTROL_ALGORITHM
    connect_timeout: float = DEFAULT_CONNECT_TIMEOUT
    connection_cleanup_interval: float = DEFAULT_CONNECTION_CLEANUP_INTERVAL
    connection_idle_check_interval: float = DEFAULT_CONNECTION_IDLE_CHECK_INTERVAL
    connection_idle_timeout: float = DEFAULT_CONNECTION_IDLE_TIMEOUT
    connection_keepalive_timeout: float = DEFAULT_CONNECTION_KEEPALIVE_TIMEOUT
    debug: bool = DEFAULT_DEBUG
    flow_control_window_auto_scale: bool = DEFAULT_FLOW_CONTROL_WINDOW_AUTO_SCALE
    flow_control_window_size: int = DEFAULT_FLOW_CONTROL_WINDOW_SIZE
    headers: Headers = field(default_factory=dict)
    initial_max_data: int = DEFAULT_INITIAL_MAX_DATA
    initial_max_streams_bidi: int = DEFAULT_INITIAL_MAX_STREAMS_BIDI
    initial_max_streams_uni: int = DEFAULT_INITIAL_MAX_STREAMS_UNI
    keep_alive: bool = DEFAULT_KEEP_ALIVE
    keyfile: str | None = None
    log_level: str = DEFAULT_LOG_LEVEL
    max_connections: int = DEFAULT_CLIENT_MAX_CONNECTIONS
    max_datagram_size: int = DEFAULT_MAX_DATAGRAM_SIZE
    max_incoming_streams: int = DEFAULT_MAX_INCOMING_STREAMS
    max_pending_events_per_session: int = DEFAULT_MAX_PENDING_EVENTS_PER_SESSION
    max_retries: int = DEFAULT_MAX_RETRIES
    max_retry_delay: float = DEFAULT_MAX_RETRY_DELAY
    max_stream_buffer_size: int = MAX_BUFFER_SIZE
    max_streams: int = DEFAULT_MAX_STREAMS
    max_total_pending_events: int = DEFAULT_MAX_TOTAL_PENDING_EVENTS
    pending_event_ttl: float = DEFAULT_PENDING_EVENT_TTL
    proxy: ProxyConfig | None = None
    read_timeout: float | None = DEFAULT_READ_TIMEOUT
    retry_backoff: float = DEFAULT_RETRY_BACKOFF
    retry_delay: float = DEFAULT_RETRY_DELAY
    stream_buffer_size: int = DEFAULT_BUFFER_SIZE
    stream_cleanup_interval: float = DEFAULT_STREAM_CLEANUP_INTERVAL
    stream_creation_timeout: float = DEFAULT_STREAM_CREATION_TIMEOUT
    stream_flow_control_increment_bidi: int = DEFAULT_STREAM_FLOW_CONTROL_INCREMENT_BIDI
    stream_flow_control_increment_uni: int = DEFAULT_STREAM_FLOW_CONTROL_INCREMENT_UNI
    user_agent: str = f"pywebtransport/{__version__}"
    verify_mode: ssl.VerifyMode | None = DEFAULT_CLIENT_VERIFY_MODE
    write_timeout: float | None = DEFAULT_WRITE_TIMEOUT

    def __post_init__(self) -> None:
        """Normalize headers and validate the configuration after initialization."""
        self.headers = normalize_headers(headers=self.headers)
        if "user-agent" not in self.headers:
            self.headers["user-agent"] = self.user_agent
        self.validate()

    @classmethod
    def create(cls, **kwargs: Any) -> Self:
        """Factory method to create a client configuration with specified overrides."""
        config_dict = {**Defaults.get_client_config(), **kwargs}
        return cls.from_dict(config_dict=config_dict)

    @classmethod
    def create_for_development(cls, *, verify_ssl: bool = False) -> Self:
        """Factory method to create a client configuration suitable for development."""
        return cls.create(
            connect_timeout=10.0,
            read_timeout=30.0,
            verify_mode=ssl.CERT_NONE if not verify_ssl else ssl.CERT_REQUIRED,
            debug=True,
            log_level="DEBUG",
        )

    @classmethod
    def create_for_production(
        cls,
        *,
        ca_certs: str | None = None,
        certfile: str | None = None,
        keyfile: str | None = None,
    ) -> Self:
        """Factory method to create a client configuration suitable for production."""
        return cls.create(
            connect_timeout=30.0,
            read_timeout=60.0,
            write_timeout=30.0,
            ca_certs=ca_certs,
            certfile=certfile,
            keyfile=keyfile,
            verify_mode=ssl.CERT_REQUIRED,
            max_streams=200,
            stream_buffer_size=65536,
            debug=False,
            log_level="INFO",
        )

    @classmethod
    def from_dict(cls, *, config_dict: dict[str, Any]) -> Self:
        """Create a ClientConfig instance from a dictionary."""
        valid_keys = {f.name for f in cls.__dataclass_fields__.values()}
        filtered_dict = {k: v for k, v in config_dict.items() if k in valid_keys}
        return cls(**filtered_dict)

    def copy(self) -> Self:
        """Create a deep copy of the configuration."""
        return copy.deepcopy(self)

    def merge(self, *, other: ClientConfig | dict[str, Any]) -> Self:
        """Merge this configuration with another config or dictionary."""
        match other:
            case dict():
                return self.update(**other)
            case ClientConfig():
                update_dict = {
                    f.name: getattr(other, f.name)
                    for f in other.__dataclass_fields__.values()
                    if getattr(other, f.name) is not None
                }
                return self.update(**update_dict)
            case _:
                raise TypeError("Can only merge with ClientConfig or dict")

    def to_dict(self) -> dict[str, Any]:
        """Convert the configuration to a dictionary."""
        result = {}
        for field_name in self.__dataclass_fields__:
            value = getattr(self, field_name)
            match value:
                case ssl.VerifyMode():
                    result[field_name] = value.name
                case _:
                    result[field_name] = value
        return result

    def update(self, **kwargs: Any) -> Self:
        """Create a new config with updated values."""
        new_config = self.copy()
        for key, value in kwargs.items():
            if hasattr(new_config, key):
                setattr(new_config, key, value)
            else:
                raise invalid_config(key=key, value=value, reason="unknown configuration key")
        new_config.validate()
        return new_config

    def validate(self) -> None:
        """Validate the integrity and correctness of the configuration values."""
        if not self.alpn_protocols:
            raise invalid_config(key="alpn_protocols", value=self.alpn_protocols, reason="cannot be empty")
        if self.congestion_control_algorithm not in SUPPORTED_CONGESTION_CONTROL_ALGORITHMS:
            raise invalid_config(
                key="congestion_control_algorithm",
                value=self.congestion_control_algorithm,
                reason=f"must be one of {SUPPORTED_CONGESTION_CONTROL_ALGORITHMS}",
            )

        try:
            _validate_timeout(timeout=self.close_timeout)
            _validate_timeout(timeout=self.connect_timeout)
            _validate_timeout(timeout=self.connection_cleanup_interval)
            _validate_timeout(timeout=self.connection_idle_check_interval)
            _validate_timeout(timeout=self.connection_idle_timeout)
            _validate_timeout(timeout=self.connection_keepalive_timeout)
            _validate_timeout(timeout=self.pending_event_ttl)
            _validate_timeout(timeout=self.read_timeout)
            _validate_timeout(timeout=self.stream_cleanup_interval)
            _validate_timeout(timeout=self.stream_creation_timeout)
            _validate_timeout(timeout=self.write_timeout)
        except ValueError as e:
            raise invalid_config(key="timeout", value=str(e), reason="invalid timeout value") from e

        if self.proxy:
            try:
                _validate_timeout(timeout=self.proxy.connect_timeout)
            except ValueError as e:
                raise invalid_config(key="proxy.connect_timeout", value=str(e), reason="invalid timeout value") from e

        if self.flow_control_window_size <= 0:
            raise invalid_config(
                key="flow_control_window_size",
                value=self.flow_control_window_size,
                reason="must be positive",
            )
        if self.max_connections <= 0:
            raise invalid_config(key="max_connections", value=self.max_connections, reason="must be positive")
        if self.max_datagram_size <= 0 or self.max_datagram_size > 65535:
            raise invalid_config(
                key="max_datagram_size",
                value=self.max_datagram_size,
                reason="must be 1-65535",
            )
        if self.max_incoming_streams <= 0:
            raise invalid_config(
                key="max_incoming_streams",
                value=self.max_incoming_streams,
                reason="must be positive",
            )
        if self.max_pending_events_per_session <= 0:
            raise invalid_config(
                key="max_pending_events_per_session",
                value=self.max_pending_events_per_session,
                reason="must be positive",
            )
        if self.max_streams <= 0:
            raise invalid_config(key="max_streams", value=self.max_streams, reason="must be positive")
        if self.max_total_pending_events <= 0:
            raise invalid_config(
                key="max_total_pending_events",
                value=self.max_total_pending_events,
                reason="must be positive",
            )
        if self.stream_buffer_size <= 0:
            raise invalid_config(
                key="stream_buffer_size",
                value=self.stream_buffer_size,
                reason="must be positive",
            )
        if self.max_stream_buffer_size < self.stream_buffer_size:
            raise invalid_config(
                key="max_stream_buffer_size",
                value=self.max_stream_buffer_size,
                reason="must be >= stream_buffer_size",
            )
        if self.stream_flow_control_increment_bidi <= 0:
            raise invalid_config(
                key="stream_flow_control_increment_bidi",
                value=self.stream_flow_control_increment_bidi,
                reason="must be positive",
            )
        if self.stream_flow_control_increment_uni <= 0:
            raise invalid_config(
                key="stream_flow_control_increment_uni",
                value=self.stream_flow_control_increment_uni,
                reason="must be positive",
            )

        if self.max_retries < 0:
            raise invalid_config(key="max_retries", value=self.max_retries, reason="must be non-negative")
        if self.max_retry_delay <= 0:
            raise invalid_config(
                key="max_retry_delay",
                value=self.max_retry_delay,
                reason="must be positive",
            )
        if self.retry_backoff < 1.0:
            raise invalid_config(key="retry_backoff", value=self.retry_backoff, reason="must be >= 1.0")
        if self.retry_delay <= 0:
            raise invalid_config(key="retry_delay", value=self.retry_delay, reason="must be positive")

        if self.ca_certs and not Path(self.ca_certs).exists():
            raise certificate_not_found(path=self.ca_certs)
        if self.certfile and not Path(self.certfile).exists():
            raise certificate_not_found(path=self.certfile)
        if self.keyfile and not Path(self.keyfile).exists():
            raise certificate_not_found(path=self.keyfile)
        if bool(self.certfile) != bool(self.keyfile):
            raise invalid_config(
                key="certfile/keyfile",
                value=f"certfile={self.certfile}, keyfile={self.keyfile}",
                reason="both must be provided together",
            )
        if self.verify_mode not in [
            ssl.CERT_NONE,
            ssl.CERT_OPTIONAL,
            ssl.CERT_REQUIRED,
            None,
        ]:
            raise invalid_config(key="verify_mode", value=self.verify_mode, reason="invalid SSL verify mode")


@dataclass(kw_only=True)
class ServerConfig:
    """A comprehensive configuration for the WebTransport server."""

    access_log: bool = DEFAULT_ACCESS_LOG
    alpn_protocols: list[str] = field(default_factory=lambda: list(DEFAULT_ALPN_PROTOCOLS))
    bind_host: str = DEFAULT_BIND_HOST
    bind_port: int = DEFAULT_DEV_PORT
    ca_certs: str | None = None
    certfile: str = DEFAULT_CERTFILE
    congestion_control_algorithm: str = DEFAULT_CONGESTION_CONTROL_ALGORITHM
    connection_cleanup_interval: float = DEFAULT_CONNECTION_CLEANUP_INTERVAL
    connection_idle_check_interval: float = DEFAULT_CONNECTION_IDLE_CHECK_INTERVAL
    connection_idle_timeout: float = DEFAULT_CONNECTION_IDLE_TIMEOUT
    connection_keepalive_timeout: float = DEFAULT_CONNECTION_KEEPALIVE_TIMEOUT
    debug: bool = DEFAULT_DEBUG
    flow_control_window_auto_scale: bool = DEFAULT_FLOW_CONTROL_WINDOW_AUTO_SCALE
    flow_control_window_size: int = DEFAULT_FLOW_CONTROL_WINDOW_SIZE
    initial_max_data: int = DEFAULT_INITIAL_MAX_DATA
    initial_max_streams_bidi: int = DEFAULT_INITIAL_MAX_STREAMS_BIDI
    initial_max_streams_uni: int = DEFAULT_INITIAL_MAX_STREAMS_UNI
    keep_alive: bool = DEFAULT_KEEP_ALIVE
    keyfile: str = DEFAULT_KEYFILE
    log_level: str = DEFAULT_LOG_LEVEL
    max_connections: int = DEFAULT_SERVER_MAX_CONNECTIONS
    max_datagram_size: int = DEFAULT_MAX_DATAGRAM_SIZE
    max_incoming_streams: int = DEFAULT_MAX_INCOMING_STREAMS
    max_pending_events_per_session: int = DEFAULT_MAX_PENDING_EVENTS_PER_SESSION
    max_sessions: int = DEFAULT_MAX_SESSIONS
    max_stream_buffer_size: int = MAX_BUFFER_SIZE
    max_streams_per_connection: int = DEFAULT_MAX_STREAMS_PER_CONNECTION
    max_total_pending_events: int = DEFAULT_MAX_TOTAL_PENDING_EVENTS
    middleware: list[MiddlewareProtocol] = field(default_factory=list)
    pending_event_ttl: float = DEFAULT_PENDING_EVENT_TTL
    read_timeout: float | None = DEFAULT_READ_TIMEOUT
    session_cleanup_interval: float = DEFAULT_SESSION_CLEANUP_INTERVAL
    stream_buffer_size: int = DEFAULT_BUFFER_SIZE
    stream_cleanup_interval: float = DEFAULT_STREAM_CLEANUP_INTERVAL
    stream_flow_control_increment_bidi: int = DEFAULT_STREAM_FLOW_CONTROL_INCREMENT_BIDI
    stream_flow_control_increment_uni: int = DEFAULT_STREAM_FLOW_CONTROL_INCREMENT_UNI
    verify_mode: ssl.VerifyMode = DEFAULT_SERVER_VERIFY_MODE
    write_timeout: float | None = DEFAULT_WRITE_TIMEOUT

    def __post_init__(self) -> None:
        """Validate the configuration after initialization."""
        self.validate()

    @classmethod
    def create(cls, **kwargs: Any) -> Self:
        """Factory method to create a server configuration with specified overrides."""
        config_dict = {**Defaults.get_server_config(), **kwargs}
        return cls.from_dict(config_dict=config_dict)

    @classmethod
    def create_for_development(
        cls,
        *,
        host: str = "localhost",
        port: int = 4433,
        certfile: str | None = None,
        keyfile: str | None = None,
    ) -> Self:
        """Factory method to create a server configuration suitable for development."""
        config = cls.create(
            bind_host=host,
            bind_port=port,
            max_connections=100,
            max_streams_per_connection=50,
            debug=True,
            log_level="DEBUG",
        )
        if certfile and keyfile:
            config.certfile, config.keyfile = certfile, keyfile
        elif not (config.certfile and config.keyfile):
            config.certfile = ""
            config.keyfile = ""
        return config

    @classmethod
    def create_for_production(
        cls,
        *,
        host: str,
        port: int,
        certfile: str,
        keyfile: str,
        ca_certs: str | None = None,
    ) -> Self:
        """Factory method to create a server configuration suitable for production."""
        return cls.create(
            bind_host=host,
            bind_port=port,
            certfile=certfile,
            keyfile=keyfile,
            ca_certs=ca_certs,
            verify_mode=ssl.CERT_OPTIONAL if ca_certs else ssl.CERT_NONE,
            max_connections=DEFAULT_SERVER_MAX_CONNECTIONS,
            max_streams_per_connection=DEFAULT_MAX_STREAMS_PER_CONNECTION,
            debug=False,
            log_level="INFO",
        )

    @classmethod
    def from_dict(cls, *, config_dict: dict[str, Any]) -> Self:
        """Create a ServerConfig instance from a dictionary."""
        valid_keys = {f.name for f in cls.__dataclass_fields__.values()}
        filtered_dict = {k: v for k, v in config_dict.items() if k in valid_keys}
        return cls(**filtered_dict)

    def copy(self) -> Self:
        """Create a deep copy of the configuration."""
        return copy.deepcopy(self)

    def get_bind_address(self) -> Address:
        """Get the bind address as a (host, port) tuple."""
        return (self.bind_host, self.bind_port)

    def merge(self, *, other: ServerConfig | dict[str, Any]) -> Self:
        """Merge this configuration with another config or dictionary."""
        match other:
            case dict():
                return self.update(**other)
            case ServerConfig():
                update_dict = {
                    f.name: getattr(other, f.name)
                    for f in other.__dataclass_fields__.values()
                    if getattr(other, f.name) is not None
                }
                return self.update(**update_dict)
            case _:
                raise TypeError("Can only merge with ServerConfig or dict")

    def to_dict(self) -> dict[str, Any]:
        """Convert the configuration to a dictionary."""
        result = {}
        for field_name in self.__dataclass_fields__:
            value = getattr(self, field_name)
            match value:
                case ssl.VerifyMode():
                    result[field_name] = value.name
                case _:
                    result[field_name] = value
        return result

    def update(self, **kwargs: Any) -> Self:
        """Create a new config with updated values."""
        new_config = self.copy()
        for key, value in kwargs.items():
            if hasattr(new_config, key):
                setattr(new_config, key, value)
            else:
                raise invalid_config(key=key, value=value, reason="unknown configuration key")
        new_config.validate()
        return new_config

    def validate(self) -> None:
        """Validate the integrity and correctness of the configuration values."""
        if not self.alpn_protocols:
            raise invalid_config(key="alpn_protocols", value=self.alpn_protocols, reason="cannot be empty")
        if not self.bind_host:
            raise invalid_config(key="bind_host", value=self.bind_host, reason="cannot be empty")
        try:
            _validate_port(port=self.bind_port)
        except ValueError as e:
            raise invalid_config(key="bind_port", value=self.bind_port, reason=str(e)) from e
        if self.congestion_control_algorithm not in SUPPORTED_CONGESTION_CONTROL_ALGORITHMS:
            raise invalid_config(
                key="congestion_control_algorithm",
                value=self.congestion_control_algorithm,
                reason=f"must be one of {SUPPORTED_CONGESTION_CONTROL_ALGORITHMS}",
            )

        try:
            _validate_timeout(timeout=self.connection_cleanup_interval)
            _validate_timeout(timeout=self.connection_idle_check_interval)
            _validate_timeout(timeout=self.connection_idle_timeout)
            _validate_timeout(timeout=self.connection_keepalive_timeout)
            _validate_timeout(timeout=self.pending_event_ttl)
            _validate_timeout(timeout=self.read_timeout)
            _validate_timeout(timeout=self.session_cleanup_interval)
            _validate_timeout(timeout=self.stream_cleanup_interval)
            _validate_timeout(timeout=self.write_timeout)
        except ValueError as e:
            raise invalid_config(key="timeout", value=str(e), reason="invalid timeout value") from e

        if self.flow_control_window_size <= 0:
            raise invalid_config(
                key="flow_control_window_size",
                value=self.flow_control_window_size,
                reason="must be positive",
            )
        if self.max_connections <= 0:
            raise invalid_config(key="max_connections", value=self.max_connections, reason="must be positive")
        if self.max_datagram_size <= 0 or self.max_datagram_size > 65535:
            raise invalid_config(
                key="max_datagram_size",
                value=self.max_datagram_size,
                reason="must be 1-65535",
            )
        if self.max_incoming_streams <= 0:
            raise invalid_config(
                key="max_incoming_streams",
                value=self.max_incoming_streams,
                reason="must be positive",
            )
        if self.max_pending_events_per_session <= 0:
            raise invalid_config(
                key="max_pending_events_per_session",
                value=self.max_pending_events_per_session,
                reason="must be positive",
            )
        if self.max_sessions <= 0:
            raise invalid_config(key="max_sessions", value=self.max_sessions, reason="must be positive")
        if self.max_stream_buffer_size < self.stream_buffer_size:
            raise invalid_config(
                key="max_stream_buffer_size",
                value=self.max_stream_buffer_size,
                reason="must be >= stream_buffer_size",
            )
        if self.max_streams_per_connection <= 0:
            raise invalid_config(
                key="max_streams_per_connection",
                value=self.max_streams_per_connection,
                reason="must be positive",
            )
        if self.max_total_pending_events <= 0:
            raise invalid_config(
                key="max_total_pending_events",
                value=self.max_total_pending_events,
                reason="must be positive",
            )
        if self.stream_buffer_size <= 0:
            raise invalid_config(
                key="stream_buffer_size",
                value=self.stream_buffer_size,
                reason="must be positive",
            )
        if self.stream_flow_control_increment_bidi <= 0:
            raise invalid_config(
                key="stream_flow_control_increment_bidi",
                value=self.stream_flow_control_increment_bidi,
                reason="must be positive",
            )
        if self.stream_flow_control_increment_uni <= 0:
            raise invalid_config(
                key="stream_flow_control_increment_uni",
                value=self.stream_flow_control_increment_uni,
                reason="must be positive",
            )

        if self.ca_certs and not Path(self.ca_certs).exists():
            raise certificate_not_found(path=self.ca_certs)
        if self.certfile and not Path(self.certfile).exists():
            raise certificate_not_found(path=self.certfile)
        if self.keyfile and not Path(self.keyfile).exists():
            raise certificate_not_found(path=self.keyfile)
        if bool(self.certfile) != bool(self.keyfile):
            raise invalid_config(
                key="certfile/keyfile",
                value=f"certfile={self.certfile}, keyfile={self.keyfile}",
                reason="certfile and keyfile must be provided together",
            )
        if self.verify_mode not in [
            ssl.CERT_NONE,
            ssl.CERT_OPTIONAL,
            ssl.CERT_REQUIRED,
        ]:
            raise invalid_config(key="verify_mode", value=self.verify_mode, reason="invalid SSL verify mode")


class ConfigBuilder:
    """A builder for fluently creating client or server configurations."""

    def __init__(self, *, config_type: str = "client") -> None:
        """Initialize the configuration builder."""
        self.config_type = config_type
        self._config_dict: dict[str, Any] = {}

    def bind(self, *, host: str, port: int) -> Self:
        """Set the bind host and port (server only)."""
        if self.config_type != "server":
            raise ConfigurationError(message="bind() can only be used with server config")

        self._config_dict["bind_host"] = host
        self._config_dict["bind_port"] = port
        return self

    def build(self) -> ClientConfig | ServerConfig:
        """Build and return the final configuration object."""
        match self.config_type:
            case "client":
                return ClientConfig.create(**self._config_dict)
            case "server":
                return ServerConfig.create(**self._config_dict)
            case _:
                raise ConfigurationError(message=f"Unknown config type: {self.config_type}")

    def debug(self, *, enabled: bool = True, log_level: str = "DEBUG") -> Self:
        """Set debug and logging settings."""
        self._config_dict["debug"] = enabled
        self._config_dict["log_level"] = log_level
        return self

    def flow_control(
        self,
        *,
        window_size: int | None = None,
        window_auto_scale: bool | None = None,
        initial_max_data: int | None = None,
        initial_max_streams_bidi: int | None = None,
        initial_max_streams_uni: int | None = None,
        stream_increment_bidi: int | None = None,
        stream_increment_uni: int | None = None,
    ) -> Self:
        """Set flow control related settings."""
        if window_size is not None:
            self._config_dict["flow_control_window_size"] = window_size
        if window_auto_scale is not None:
            self._config_dict["flow_control_window_auto_scale"] = window_auto_scale
        if initial_max_data is not None:
            self._config_dict["initial_max_data"] = initial_max_data
        if initial_max_streams_bidi is not None:
            self._config_dict["initial_max_streams_bidi"] = initial_max_streams_bidi
        if initial_max_streams_uni is not None:
            self._config_dict["initial_max_streams_uni"] = initial_max_streams_uni
        if stream_increment_bidi is not None:
            self._config_dict["stream_flow_control_increment_bidi"] = stream_increment_bidi
        if stream_increment_uni is not None:
            self._config_dict["stream_flow_control_increment_uni"] = stream_increment_uni
        return self

    def performance(
        self,
        *,
        max_streams: int | None = None,
        buffer_size: int | None = None,
        max_connections: int | None = None,
    ) -> Self:
        """Set performance-related settings."""
        if max_streams is not None:
            if self.config_type == "client":
                self._config_dict["max_streams"] = max_streams
            else:
                self._config_dict["max_streams_per_connection"] = max_streams
        if buffer_size is not None:
            self._config_dict["stream_buffer_size"] = buffer_size
        if max_connections is not None and self.config_type == "server":
            self._config_dict["max_connections"] = max_connections
        return self

    def security(
        self,
        *,
        certfile: str | None = None,
        keyfile: str | None = None,
        ca_certs: str | None = None,
        verify_mode: ssl.VerifyMode | None = None,
    ) -> Self:
        """Set security and SSL/TLS settings."""
        if certfile is not None:
            self._config_dict["certfile"] = certfile
        if keyfile is not None:
            self._config_dict["keyfile"] = keyfile
        if ca_certs is not None:
            self._config_dict["ca_certs"] = ca_certs
        if verify_mode is not None:
            self._config_dict["verify_mode"] = verify_mode
        return self

    def timeout(
        self,
        *,
        connect: float | None = None,
        read: float | None = None,
        write: float | None = None,
    ) -> Self:
        """Set timeout values for the configuration."""
        if connect is not None:
            self._config_dict["connect_timeout"] = connect
        if read is not None:
            self._config_dict["read_timeout"] = read
        if write is not None:
            self._config_dict["write_timeout"] = write
        return self


def _validate_port(*, port: Any) -> None:
    """Validate that a value is a valid network port."""
    if not isinstance(port, int) or not (1 <= port <= 65535):
        raise ValueError(f"Port must be an integer between 1 and 65535, got {port}")


def _validate_timeout(*, timeout: float | None) -> None:
    """Validate a timeout value."""
    if timeout is not None:
        if not isinstance(timeout, (int, float)):
            raise TypeError("Timeout must be a number or None")
        if timeout <= 0:
            raise ValueError("Timeout must be positive")



"""WebTransport connection utility functions."""

from __future__ import annotations

import asyncio

from pywebtransport.config import ClientConfig
from pywebtransport.connection.connection import WebTransportConnection
from pywebtransport.exceptions import ConnectionError, HandshakeError
from pywebtransport.utils import get_logger

__all__ = [
    "connect_with_retry",
    "create_multiple_connections",
    "ensure_connection",
    "test_multiple_connections",
    "test_tcp_connection",
]

logger = get_logger(name="connection.utils")


async def connect_with_retry(
    *,
    config: ClientConfig,
    host: str,
    port: int,
    path: str = "/",
    max_retries: int = 3,
    retry_delay: float = 1.0,
    backoff_factor: float = 2.0,
) -> WebTransportConnection:
    """Establish a connection with an exponential backoff retry mechanism."""
    last_error: Exception | None = None
    for attempt in range(max_retries + 1):
        try:
            connection = await WebTransportConnection.create_client(config=config, host=host, port=port, path=path)
            if attempt > 0:
                logger.info("Connected to %s:%s after %d retries", host, port, attempt)
            return connection
        except (ConnectionError, HandshakeError) as e:
            last_error = e
            if attempt < max_retries:
                delay = retry_delay * (backoff_factor**attempt)
                logger.warning(
                    "Connection attempt %d failed, retrying in %.1fs: %s",
                    attempt + 1,
                    delay,
                    e,
                    exc_info=True,
                )
                await asyncio.sleep(delay)
            else:
                logger.error("All %d connection attempts failed", max_retries + 1)
    raise ConnectionError(message=f"Failed to connect after {max_retries + 1} attempts: {last_error}")


async def create_multiple_connections(
    *,
    config: ClientConfig,
    targets: list[tuple[str, int]],
    path: str = "/",
    max_concurrent: int = 10,
) -> dict[str, WebTransportConnection]:
    """Create multiple connections to a list of targets with a concurrency limit."""
    semaphore = asyncio.Semaphore(value=max_concurrent)
    connections: dict[str, WebTransportConnection] = {}

    async def create_single_connection(*, host: str, port: int) -> None:
        target_key = f"{host}:{port}"
        async with semaphore:
            try:
                connection = await WebTransportConnection.create_client(config=config, host=host, port=port, path=path)
                connections[target_key] = connection
            except Exception as e:
                logger.error("Failed to connect to %s:%s: %s", host, port, e, exc_info=True)

    try:
        async with asyncio.TaskGroup() as tg:
            for host, port in targets:
                tg.create_task(create_single_connection(host=host, port=port))
    except* Exception as eg:
        logger.error(
            "Errors occurred while creating multiple connections: %s",
            eg.exceptions,
            exc_info=eg,
        )

    return connections


async def ensure_connection(
    *,
    connection: WebTransportConnection,
    config: ClientConfig,
    host: str,
    port: int,
    path: str = "/",
    reconnect: bool = True,
) -> WebTransportConnection:
    """Ensure a connection is active, optionally reconnecting if it is not."""
    if connection.is_connected:
        return connection
    if not reconnect:
        raise ConnectionError(message="Connection not active and reconnect disabled")

    logger.info("Reconnecting to %s:%s", host, port)
    await connection.close()
    new_connection = await WebTransportConnection.create_client(config=config, host=host, port=port, path=path)
    return new_connection


async def test_multiple_connections(*, targets: list[tuple[str, int]], timeout: float = 10.0) -> dict[str, bool]:
    """Test TCP connectivity to multiple targets concurrently."""
    connection_results: dict[str, bool] = {}
    tasks: dict[str, asyncio.Task[bool]] = {}

    try:
        async with asyncio.TaskGroup() as tg:
            for host, port in targets:
                target_key = f"{host}:{port}"
                tasks[target_key] = tg.create_task(test_tcp_connection(host=host, port=port, timeout=timeout))
    except* Exception:
        pass

    for target_key, task in tasks.items():
        if task.done() and not task.cancelled():
            try:
                connection_results[target_key] = task.result()
            except Exception:
                connection_results[target_key] = False
        else:
            connection_results[target_key] = False

    return connection_results


async def test_tcp_connection(*, host: str, port: int, timeout: float = 10.0) -> bool:
    """Test if a TCP connection can be established to a host and port."""
    try:
        _, writer = await asyncio.wait_for(asyncio.open_connection(host=host, port=port), timeout=timeout)
        writer.close()
        await writer.wait_closed()
        return True
    except (asyncio.TimeoutError, OSError):
        return False



"""WebTransport Browser Client."""

from __future__ import annotations

import asyncio
from types import TracebackType
from typing import Self

from pywebtransport.client.client import WebTransportClient
from pywebtransport.config import ClientConfig
from pywebtransport.exceptions import ClientError
from pywebtransport.session import WebTransportSession
from pywebtransport.utils import get_logger

__all__ = ["WebTransportBrowser"]

logger = get_logger(name="client.browser")


class WebTransportBrowser:
    """A browser-like WebTransport client with a managed lifecycle and history."""

    def __init__(self, *, config: ClientConfig | None = None) -> None:
        """Initialize the browser-like WebTransport client."""
        self._client = WebTransportClient(config=config)
        self._history: list[str] = []
        self._history_index: int = -1
        self._current_session: WebTransportSession | None = None
        self._lock: asyncio.Lock | None = None

    @classmethod
    def create(cls, *, config: ClientConfig | None = None) -> Self:
        """Factory method to create a new browser-like client instance."""
        return cls(config=config)

    @property
    def current_session(self) -> WebTransportSession | None:
        """Get the current active session."""
        return self._current_session

    async def __aenter__(self) -> Self:
        """Enter the async context, initializing resources."""
        self._lock = asyncio.Lock()
        await self._client.__aenter__()
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        """Exit the async context, ensuring all resources are closed."""
        await self.close()

    async def close(self) -> None:
        """Close the browser, the current session, and all underlying resources."""
        if self._lock is None:
            raise ClientError(
                message=(
                    "WebTransportBrowser has not been activated. It must be used as an "
                    "asynchronous context manager (`async with ...`)."
                )
            )

        async with self._lock:
            logger.info("Closing WebTransportBrowser and active session.")
            if self._current_session and not self._current_session.is_closed:
                await self._current_session.close()

            await self._client.close()
            self._current_session = None
            self._history.clear()
            self._history_index = -1

    async def back(self) -> WebTransportSession | None:
        """Go back to the previous entry in the navigation history."""
        if self._lock is None:
            raise ClientError(
                message=(
                    "WebTransportBrowser has not been activated. It must be used as an "
                    "asynchronous context manager (`async with ...`)."
                )
            )

        async with self._lock:
            if self._history_index > 0:
                self._history_index -= 1
                url_to_navigate = self._history[self._history_index]
                return await self._navigate_internal(url=url_to_navigate)
            return None

    async def forward(self) -> WebTransportSession | None:
        """Go forward to the next entry in the navigation history."""
        if self._lock is None:
            raise ClientError(
                message=(
                    "WebTransportBrowser has not been activated. It must be used as an "
                    "asynchronous context manager (`async with ...`)."
                )
            )

        async with self._lock:
            if self._history_index < len(self._history) - 1:
                self._history_index += 1
                url_to_navigate = self._history[self._history_index]
                return await self._navigate_internal(url=url_to_navigate)
            return None

    async def navigate(self, *, url: str) -> WebTransportSession:
        """Navigate to a URL, creating a new session and clearing forward history."""
        if self._lock is None:
            raise ClientError(
                message=(
                    "WebTransportBrowser has not been activated. It must be used as an "
                    "asynchronous context manager (`async with ...`)."
                )
            )

        async with self._lock:
            if self._current_session and self._history and self._history[self._history_index] == url:
                return self._current_session

            if self._history_index < len(self._history) - 1:
                self._history = self._history[: self._history_index + 1]

            if not self._history or self._history[-1] != url:
                self._history.append(url)

            self._history_index = len(self._history) - 1
            return await self._navigate_internal(url=url)

    async def refresh(self) -> WebTransportSession | None:
        """Refresh the current session by reconnecting to the current URL."""
        if self._lock is None:
            raise ClientError(
                message=(
                    "WebTransportBrowser has not been activated. It must be used as an "
                    "asynchronous context manager (`async with ...`)."
                )
            )

        async with self._lock:
            if not self._history:
                return None
            current_url = self._history[self._history_index]
            return await self._navigate_internal(url=current_url)

    async def get_history(self) -> list[str]:
        """Get a copy of the navigation history."""
        if self._lock is None:
            raise ClientError(
                message=(
                    "WebTransportBrowser has not been activated. It must be used as an "
                    "asynchronous context manager (`async with ...`)."
                )
            )

        async with self._lock:
            return self._history.copy()

    async def _navigate_internal(self, *, url: str) -> WebTransportSession:
        """Handle the core navigation logic for session teardown and creation."""
        old_session = self._current_session
        self._current_session = None

        if old_session and not old_session.is_closed:
            asyncio.create_task(old_session.close())

        try:
            new_session = await self._client.connect(url=url)
            self._current_session = new_session
            return new_session
        except Exception as e:
            self._current_session = old_session
            logger.error("Failed to navigate to %s: %s", url, e, exc_info=True)
            raise



"""WebTransport Pub/Sub-specific exceptions."""

from __future__ import annotations

from pywebtransport.exceptions import WebTransportError

__all__ = [
    "NotSubscribedError",
    "PubSubError",
    "SubscriptionFailedError",
]


class PubSubError(WebTransportError):
    """Base exception for Pub/Sub errors."""

    pass


class NotSubscribedError(PubSubError):
    """Raised when trying to operate on a topic without being subscribed."""

    pass


class SubscriptionFailedError(PubSubError):
    """Raised when subscribing to a topic fails."""

    pass



"""WebTransport Datagram Performance Monitor."""

from __future__ import annotations

import asyncio
from collections import deque
from types import TracebackType
from typing import TYPE_CHECKING, Any, Self

from pywebtransport.utils import get_logger, get_timestamp

if TYPE_CHECKING:
    from pywebtransport.datagram.transport import WebTransportDatagramTransport


__all__ = ["DatagramMonitor"]

logger = get_logger(name="datagram.monitor")


class DatagramMonitor:
    """Monitors datagram transport performance and generates alerts."""

    def __init__(
        self,
        datagram_transport: WebTransportDatagramTransport,
        *,
        monitoring_interval: float = 5.0,
        samples_maxlen: int = 100,
        alerts_maxlen: int = 50,
        queue_size_threshold: float = 0.9,
        success_rate_threshold: float = 0.8,
        trend_analysis_window: int = 10,
    ) -> None:
        """Initialize the datagram performance monitor."""
        self._transport = datagram_transport
        self._interval = monitoring_interval
        self._monitor_task: asyncio.Task[None] | None = None
        self._samples: deque[dict[str, Any]] = deque(maxlen=samples_maxlen)
        self._alerts: deque[dict[str, Any]] = deque(maxlen=alerts_maxlen)
        self._queue_size_threshold = queue_size_threshold
        self._success_rate_threshold = success_rate_threshold
        self._trend_analysis_window = trend_analysis_window

    @classmethod
    def create(
        cls,
        *,
        datagram_transport: WebTransportDatagramTransport,
        monitoring_interval: float = 5.0,
    ) -> Self:
        """Factory method to create a new datagram monitor for a transport."""
        return cls(datagram_transport=datagram_transport, monitoring_interval=monitoring_interval)

    @property
    def is_monitoring(self) -> bool:
        """Check if the monitoring task is currently active."""
        return self._monitor_task is not None and not self._monitor_task.done()

    async def __aenter__(self) -> Self:
        """Enter the async context and start the monitoring task."""
        if self.is_monitoring:
            return self

        try:
            self._monitor_task = asyncio.create_task(self._monitor_loop())
            logger.info(
                "Datagram monitoring started for session %s...",
                self._transport.session_id[:12],
            )
        except RuntimeError:
            logger.error(
                "Failed to start datagram monitor: No running event loop.",
                exc_info=True,
            )

        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        """Exit the async context and stop the monitoring task."""
        if self._monitor_task:
            self._monitor_task.cancel()
            try:
                await self._monitor_task
            except asyncio.CancelledError:
                pass
        logger.info(
            "Datagram monitoring stopped for session %s...",
            self._transport.session_id[:12],
        )

    def get_alerts(self) -> list[dict[str, Any]]:
        """Get a copy of the currently active alerts."""
        return list(self._alerts)

    def get_samples(self, *, limit: int | None = None) -> list[dict[str, Any]]:
        """Get a copy of the collected performance samples."""
        samples_list = list(self._samples)
        if limit is not None:
            return samples_list[-limit:]
        return samples_list

    def _analyze_trend(self, *, values: list[float]) -> str:
        """Perform a simple trend analysis on a series of values."""
        if len(values) < self._trend_analysis_window:
            return "stable"

        first_half = values[: len(values) // 2]
        second_half = values[len(values) // 2 :]

        if not first_half or not second_half:
            return "stable"

        first_avg = sum(first_half) / len(first_half)
        second_avg = sum(second_half) / len(second_half)

        if first_avg == 0 and second_avg > 0:
            return "increasing"

        if first_avg > 0:
            change = (second_avg - first_avg) / first_avg
            if change > 0.25:
                return "increasing"
            elif change < -0.25:
                return "decreasing"

        return "stable"

    async def _check_alerts(self, *, current_sample: dict[str, Any]) -> None:
        """Check the current sample against configured alert thresholds."""
        if (
            current_sample["outgoing_queue_size"]
            > self._transport.outgoing_high_water_mark * self._queue_size_threshold
        ):
            self._alerts.append(
                {
                    "type": "high_queue_size",
                    "message": f"Outgoing queue size high: {current_sample['outgoing_queue_size']}",
                    "timestamp": current_sample["timestamp"],
                }
            )

        if current_sample["send_success_rate"] < self._success_rate_threshold:
            self._alerts.append(
                {
                    "type": "low_success_rate",
                    "message": f"Low send success rate: {current_sample['send_success_rate']:.2%}",
                    "timestamp": current_sample["timestamp"],
                }
            )

        if len(self._samples) >= self._trend_analysis_window:
            recent_send_times = [s["avg_send_time"] for s in self._samples]
            trend = self._analyze_trend(values=recent_send_times)
            if trend == "increasing":
                self._alerts.append(
                    {
                        "type": "increasing_send_time",
                        "message": f"Average send time is increasing (current: {current_sample['avg_send_time']:.3f}s)",
                        "timestamp": current_sample["timestamp"],
                    }
                )

    async def _monitor_loop(self) -> None:
        """The main loop for collecting stats and checking for alerts."""
        try:
            while True:
                await asyncio.sleep(self._interval)
                stats = self._transport.stats
                queue_stats = self._transport.get_queue_stats()

                sample = {
                    "timestamp": get_timestamp(),
                    "datagrams_sent": stats.get("datagrams_sent", 0),
                    "datagrams_received": stats.get("datagrams_received", 0),
                    "send_success_rate": stats.get("send_success_rate", 0.0),
                    "avg_send_time": stats.get("avg_send_time", 0.0),
                    "outgoing_queue_size": queue_stats.get("outgoing", {}).get("size", 0),
                }

                self._samples.append(sample)
                await self._check_alerts(current_sample=sample)

        except asyncio.CancelledError:
            logger.info("Datagram monitor loop cancelled.")
        except Exception as e:
            logger.error("Monitor loop error: %s", e, exc_info=e)



"""Asynchronous Event System."""

from __future__ import annotations

import asyncio
import uuid
from collections import defaultdict, deque
from collections.abc import Callable
from dataclasses import dataclass, field
from typing import Any, Self, TypeVar

from pywebtransport.types import EventData, EventHandler, EventType, Timeout
from pywebtransport.utils import get_logger, get_timestamp

__all__ = [
    "Event",
    "EventBus",
    "EventEmitter",
    "create_event_bus",
    "create_event_emitter",
    "event_handler",
]

logger = get_logger(name="events")
F = TypeVar("F", bound=Callable[..., Any])


@dataclass(kw_only=True)
class Event:
    """A versatile base class for all system events."""

    type: EventType | str
    timestamp: float = field(default_factory=get_timestamp)
    data: EventData | None = None
    source: Any | None = None
    event_id: str = field(default_factory=lambda: str(uuid.uuid4()))

    def __post_init__(self) -> None:
        """Handle string-based event types after initialization."""
        if isinstance(self.type, str):
            try:
                self.type = EventType(self.type)
            except ValueError:
                pass

    @classmethod
    def for_connection(cls, *, event_type: EventType, connection_info: dict[str, Any]) -> Self:
        """Factory method to create a new connection event."""
        return cls(type=event_type, data=connection_info)

    @classmethod
    def for_datagram(cls, *, event_type: EventType, datagram_info: dict[str, Any]) -> Self:
        """Factory method to create a new datagram event."""
        return cls(type=event_type, data=datagram_info)

    @classmethod
    def for_error(cls, *, error: Exception, source: Any = None) -> Self:
        """Factory method to create a new error event from an exception."""
        to_dict_method: Callable[[], dict[str, Any]] = getattr(error, "to_dict", lambda: {})
        details = to_dict_method() if callable(to_dict_method) else {}
        return cls(
            type=EventType.PROTOCOL_ERROR,
            data={
                "error_type": type(error).__name__,
                "error_message": str(error),
                "error_details": details,
            },
            source=source,
        )

    @classmethod
    def for_session(cls, *, event_type: EventType, session_info: dict[str, Any]) -> Self:
        """Factory method to create a new session event."""
        return cls(type=event_type, data=session_info)

    @classmethod
    def for_stream(cls, *, event_type: EventType, stream_info: dict[str, Any]) -> Self:
        """Factory method to create a new stream event."""
        return cls(type=event_type, data=stream_info)

    @property
    def is_connection_event(self) -> bool:
        """Check if this event is connection-related."""
        return self.type.startswith("connection_")

    @property
    def is_datagram_event(self) -> bool:
        """Check if this event is datagram-related."""
        return self.type.startswith("datagram_")

    @property
    def is_error_event(self) -> bool:
        """Check if this event is error-related."""
        return "error" in self.type.lower()

    @property
    def is_session_event(self) -> bool:
        """Check if this event is session-related."""
        return self.type.startswith("session_")

    @property
    def is_stream_event(self) -> bool:
        """Check if this event is stream-related."""
        return self.type.startswith("stream_")

    def to_dict(self) -> dict[str, Any]:
        """Convert the event to a dictionary."""
        return {
            "id": self.event_id,
            "type": self.type,
            "timestamp": self.timestamp,
            "data": self.data,
            "source": str(self.source) if self.source else None,
        }

    def __str__(self) -> str:
        """Return a simple string representation of the event."""
        return f"Event({self.type}, {self.event_id[:8]})"

    def __repr__(self) -> str:
        """Return a detailed string representation of the event."""
        return f"Event(type={self.type}, id={self.event_id}, timestamp={self.timestamp})"


class EventEmitter:
    """An emitter for handling and dispatching events asynchronously."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initialize the event emitter."""
        if getattr(self, "_emitter_initialized", False):
            return
        self._handlers: dict[EventType | str, list[EventHandler]] = defaultdict(list)
        self._once_handlers: dict[EventType | str, list[EventHandler]] = defaultdict(list)
        self._max_listeners = kwargs.get("max_listeners", 100)
        self._event_history: list[Event] = []
        self._max_history = 1000
        self._wildcard_handlers: list[EventHandler] = []
        self._paused = False
        self._event_queue: deque[Event] = deque()
        self._processing_task: asyncio.Task | None = None
        self._emitter_initialized = True

    async def close(self) -> None:
        """Cancel running event processing tasks and clear all listeners."""
        if self._processing_task and not self._processing_task.done():
            self._processing_task.cancel()
            try:
                await self._processing_task
            except asyncio.CancelledError:
                pass
        self.remove_all_listeners()
        logger.debug("EventEmitter closed and listeners cleared.")

    def pause(self) -> None:
        """Pause event processing and queue subsequent events."""
        self._paused = True
        logger.debug("Event processing paused")

    def resume(self) -> asyncio.Task | None:
        """Resume event processing and handle all queued events."""
        self._paused = False
        logger.debug("Event processing resumed")
        if self._event_queue and (self._processing_task is None or self._processing_task.done()):
            self._processing_task = asyncio.create_task(self._process_queued_events())
            return self._processing_task
        return None

    async def emit(
        self,
        *,
        event_type: EventType | str,
        data: EventData | None = None,
        source: Any = None,
    ) -> None:
        """Emit an event to all corresponding listeners."""
        event = Event(type=event_type, data=data, source=source)
        self._add_to_history(event=event)

        if self._paused:
            self._event_queue.append(event)
            return

        await self._process_event(event=event)

    def off(self, *, event_type: EventType | str, handler: EventHandler | None = None) -> None:
        """Unregister a specific event handler or all handlers for an event."""
        if handler is None:
            self._handlers[event_type].clear()
            self._once_handlers[event_type].clear()
            logger.debug("Removed all handlers for event %s", event_type)
        else:
            if handler in self._handlers[event_type]:
                self._handlers[event_type].remove(handler)
                logger.debug("Removed handler for event %s", event_type)
            if handler in self._once_handlers[event_type]:
                self._once_handlers[event_type].remove(handler)
                logger.debug("Removed once handler for event %s", event_type)

    def off_any(self, *, handler: EventHandler | None = None) -> None:
        """Unregister a specific wildcard handler or all wildcard handlers."""
        if handler is None:
            self._wildcard_handlers.clear()
            logger.debug("Removed all wildcard handlers")
        elif handler in self._wildcard_handlers:
            self._wildcard_handlers.remove(handler)
            logger.debug("Removed wildcard handler")

    def on(self, *, event_type: EventType | str, handler: EventHandler) -> None:
        """Register a persistent event handler."""
        handlers = self._handlers[event_type]
        if len(handlers) >= self._max_listeners:
            logger.warning(
                "Maximum listeners (%d) exceeded for event %s",
                self._max_listeners,
                event_type,
            )
        if handler not in handlers:
            handlers.append(handler)
            logger.debug("Registered handler for event %s", event_type)
        else:
            logger.warning("Handler already registered for event %s", event_type)

    def on_any(self, *, handler: EventHandler) -> None:
        """Register a wildcard handler for all events."""
        if handler not in self._wildcard_handlers:
            self._wildcard_handlers.append(handler)
            logger.debug("Registered wildcard handler")

    def once(self, *, event_type: EventType | str, handler: EventHandler) -> None:
        """Register a one-time event handler."""
        once_handlers = self._once_handlers[event_type]
        if handler not in once_handlers:
            once_handlers.append(handler)
            logger.debug("Registered once handler for event %s", event_type)

    async def wait_for(
        self,
        *,
        event_type: EventType | str,
        timeout: Timeout | None = None,
        condition: Callable[[Event], bool] | None = None,
    ) -> Event:
        """Wait for a specific event to be emitted."""
        future: asyncio.Future[Event] = asyncio.Future()

        async def handler(event: Event) -> None:
            try:
                if condition is None or condition(event):
                    if not future.done():
                        future.set_result(event)
            except Exception as e:
                if not future.done():
                    future.set_exception(e)

        self.on(event_type=event_type, handler=handler)
        try:
            return await asyncio.wait_for(future, timeout=timeout)
        finally:
            self.off(event_type=event_type, handler=handler)

    def clear_history(self) -> None:
        """Clear the entire event history."""
        self._event_history.clear()
        logger.debug("Event history cleared")

    def get_event_history(self, *, event_type: EventType | str | None = None, limit: int = 100) -> list[Event]:
        """Get the recorded history of events."""
        if event_type is None:
            return self._event_history[-limit:]

        filtered_events = [event for event in self._event_history if event.type == event_type]
        return filtered_events[-limit:]

    def get_stats(self) -> dict[str, Any]:
        """Get statistics about the event emitter."""
        total_handlers = sum(len(handlers) for handlers in self._handlers.values())
        total_once_handlers = sum(len(handlers) for handlers in self._once_handlers.values())
        return {
            "total_handlers": total_handlers,
            "total_once_handlers": total_once_handlers,
            "wildcard_handlers": len(self._wildcard_handlers),
            "event_types": len(self._handlers),
            "history_size": len(self._event_history),
            "queued_events": len(self._event_queue),
            "paused": self._paused,
        }

    def listener_count(self, *, event_type: EventType | str) -> int:
        """Get the number of listeners for a specific event type."""
        return len(self.listeners(event_type=event_type))

    def listeners(self, *, event_type: EventType | str) -> list[EventHandler]:
        """Get all listeners for a specific event type."""
        return self._handlers[event_type][:] + self._once_handlers[event_type][:]

    def remove_all_listeners(self, *, event_type: EventType | str | None = None) -> None:
        """Remove all listeners for a specific event or for all events."""
        if event_type is None:
            self._handlers.clear()
            self._once_handlers.clear()
            self._wildcard_handlers.clear()
            logger.debug("Removed all event listeners")
        else:
            self._handlers[event_type].clear()
            self._once_handlers[event_type].clear()
            logger.debug("Removed all listeners for event %s", event_type)

    def set_max_listeners(self, *, max_listeners: int) -> None:
        """Set the maximum number of listeners per event."""
        self._max_listeners = max_listeners

    def _add_to_history(self, *, event: Event) -> None:
        """Add an event to the history buffer."""
        self._event_history.append(event)
        if len(self._event_history) > self._max_history:
            self._event_history.pop(0)

    async def _process_event(self, *, event: Event) -> None:
        """Process a single event by invoking all relevant handlers."""
        handlers_to_call: list[EventHandler] = self._handlers[event.type][:]
        once_handlers_to_call: list[EventHandler] = self._once_handlers[event.type][:]
        if once_handlers_to_call:
            self._once_handlers[event.type].clear()
        all_handlers = handlers_to_call + once_handlers_to_call + self._wildcard_handlers
        if not all_handlers:
            return
        logger.debug("Emitting event %s to %d handlers", event.type, len(all_handlers))
        for handler in all_handlers:
            try:
                if asyncio.iscoroutinefunction(handler):
                    await handler(event)
                else:
                    handler(event)
            except Exception as e:
                logger.error("Error in handler for event %s: %s", event.type, e, exc_info=True)

    async def _process_queued_events(self) -> None:
        """Process all events in the queue until it is empty."""
        while self._event_queue and not self._paused:
            event = self._event_queue.popleft()
            await self._process_event(event=event)


class EventBus:
    """A global, singleton event bus for cross-component communication."""

    _instance: EventBus | None = None
    _lock: asyncio.Lock | None = None

    def __init__(self) -> None:
        """Initialize the event bus."""
        self._emitter = EventEmitter(max_listeners=1000)
        self._subscriptions: dict[str, tuple[EventType | str, EventHandler]] = {}
        self._subscription_counter = 0

    @classmethod
    async def get_instance(cls) -> EventBus:
        """Get the singleton instance of the EventBus."""
        if cls._instance is None:
            if cls._lock is None:
                cls._lock = asyncio.Lock()
            async with cls._lock:
                if cls._instance is None:
                    cls._instance = cls()
        return cls._instance

    async def close(self) -> None:
        """Close the event bus and its underlying emitter."""
        await self._emitter.close()
        self.clear_all_subscriptions()
        logger.debug("Event bus closed.")

    async def emit(
        self,
        *,
        event_type: EventType | str,
        data: EventData | None = None,
        source: Any = None,
    ) -> None:
        """Create and emit an event on the bus."""
        await self._emitter.emit(event_type=event_type, data=data, source=source)

    async def publish(self, *, event: Event) -> None:
        """Publish a pre-constructed event to all subscribers."""
        await self._emitter.emit(event_type=event.type, data=event.data, source=event.source)

    def subscribe(self, *, event_type: EventType | str, handler: EventHandler, once: bool = False) -> str:
        """Subscribe to an event, returning a unique subscription ID."""
        subscription_id = f"sub_{self._subscription_counter}"
        self._subscription_counter += 1
        self._subscriptions[subscription_id] = (event_type, handler)
        if once:
            self._emitter.once(event_type=event_type, handler=handler)
        else:
            self._emitter.on(event_type=event_type, handler=handler)
        logger.debug("Created subscription %s for event %s", subscription_id, event_type)
        return subscription_id

    def unsubscribe(self, *, subscription_id: str) -> None:
        """Unsubscribe from an event using its subscription ID."""
        if subscription_id not in self._subscriptions:
            logger.warning("Subscription %s not found", subscription_id)
            return
        event_type, handler = self._subscriptions.pop(subscription_id)
        self._emitter.off(event_type=event_type, handler=handler)
        logger.debug("Removed subscription %s for event %s", subscription_id, event_type)

    def clear_all_subscriptions(self) -> None:
        """Clear all subscriptions from the bus."""
        self._subscriptions.clear()
        self._emitter.remove_all_listeners()
        logger.debug("Cleared all event bus subscriptions")

    def get_subscription_count(self) -> int:
        """Get the number of active subscriptions."""
        return len(self._subscriptions)


async def create_event_bus() -> EventBus:
    """Create or get the global event bus instance."""
    return await EventBus.get_instance()


def create_event_emitter(*, max_listeners: int = 100) -> EventEmitter:
    """Create a new, standalone event emitter."""
    return EventEmitter(max_listeners=max_listeners)


def event_handler(*, event_type: EventType | str) -> Callable[[F], F]:
    """Decorate a function as a handler for a specific event type."""

    def decorator(func: F) -> F:
        setattr(func, "_event_type", event_type)
        setattr(func, "_is_event_handler", True)
        return func

    return decorator



"""WebTransport session manager implementation."""

from __future__ import annotations

import asyncio
import weakref
from collections import defaultdict
from types import TracebackType
from typing import Any, Self

from pywebtransport.constants import DEFAULT_MAX_SESSIONS, DEFAULT_SESSION_CLEANUP_INTERVAL
from pywebtransport.exceptions import SessionError
from pywebtransport.session.session import WebTransportSession
from pywebtransport.types import EventType, SessionId, SessionState
from pywebtransport.utils import get_logger

__all__ = ["SessionManager"]

logger = get_logger(name="session.manager")


class SessionManager:
    """Manages multiple WebTransport sessions with concurrency safety."""

    def __init__(
        self,
        *,
        max_sessions: int = DEFAULT_MAX_SESSIONS,
        session_cleanup_interval: float = DEFAULT_SESSION_CLEANUP_INTERVAL,
    ) -> None:
        """Initialize the session manager."""
        self._max_sessions = max_sessions
        self._cleanup_interval = session_cleanup_interval
        self._lock: asyncio.Lock | None = None
        self._sessions: dict[SessionId, WebTransportSession] = {}
        self._stats = {
            "total_created": 0,
            "total_closed": 0,
            "current_count": 0,
            "max_concurrent": 0,
        }
        self._cleanup_task: asyncio.Task[None] | None = None
        self._is_shutting_down = False

    @classmethod
    def create(
        cls,
        *,
        max_sessions: int = DEFAULT_MAX_SESSIONS,
        session_cleanup_interval: float = DEFAULT_SESSION_CLEANUP_INTERVAL,
    ) -> Self:
        """Factory method to create a new session manager instance."""
        return cls(max_sessions=max_sessions, session_cleanup_interval=session_cleanup_interval)

    async def __aenter__(self) -> Self:
        """Enter async context, initializing resources and starting background tasks."""
        self._lock = asyncio.Lock()
        self._start_cleanup_task()
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        """Exit async context, shutting down the manager."""
        await self.shutdown()

    async def shutdown(self) -> None:
        """Shut down the session manager and close all active sessions."""
        if self._is_shutting_down:
            return

        self._is_shutting_down = True
        logger.info("Shutting down session manager")

        if self._cleanup_task:
            self._cleanup_task.cancel()
            try:
                await self._cleanup_task
            except asyncio.CancelledError:
                pass

        await self.close_all_sessions()
        logger.info("Session manager shutdown complete")

    async def add_session(self, *, session: WebTransportSession) -> SessionId:
        """Add a new session to the manager."""
        if self._lock is None:
            raise SessionError(
                message=(
                    "SessionManager has not been activated. It must be used as an "
                    "asynchronous context manager (`async with ...`)."
                )
            )

        async with self._lock:
            if len(self._sessions) >= self._max_sessions:
                raise SessionError(message=f"Maximum sessions ({self._max_sessions}) exceeded")

            session_id = session.session_id
            self._sessions[session_id] = session
            manager_ref = weakref.ref(self)

            async def on_close(event: Any) -> None:
                manager = manager_ref()
                if manager and isinstance(event.data, dict):
                    await manager.remove_session(session_id=event.data["session_id"])

            session.once(event_type=EventType.SESSION_CLOSED, handler=on_close)

            self._stats["total_created"] += 1
            self._update_stats_unsafe()
            logger.debug("Added session %s (total: %d)", session_id, len(self._sessions))
            return session_id

    async def close_all_sessions(self) -> None:
        """Close and remove all currently managed sessions."""
        if self._lock is None:
            raise SessionError(
                message=(
                    "SessionManager has not been activated. It must be used as an "
                    "asynchronous context manager (`async with ...`)."
                )
            )

        sessions_to_close: list[WebTransportSession] = []
        async with self._lock:
            if not self._sessions:
                return

            sessions_to_close = list(self._sessions.values())
            logger.info("Initiating shutdown for %d managed sessions.", len(sessions_to_close))
            self._stats["total_closed"] += len(sessions_to_close)
            self._sessions.clear()
            self._update_stats_unsafe()

        try:
            async with asyncio.TaskGroup() as tg:
                for session in sessions_to_close:
                    if not session.is_closed:
                        tg.create_task(session.close(close_connection=False))
        except* Exception as eg:
            logger.error("Errors occurred while closing managed sessions: %s", eg.exceptions, exc_info=eg)
            raise

    async def get_session(self, *, session_id: SessionId) -> WebTransportSession | None:
        """Retrieve a session by its ID."""
        if self._lock is None:
            raise SessionError(
                message=(
                    "SessionManager has not been activated. It must be used as an "
                    "asynchronous context manager (`async with ...`)."
                )
            )

        async with self._lock:
            return self._sessions.get(session_id)

    async def remove_session(self, *, session_id: SessionId) -> WebTransportSession | None:
        """Remove a session from the manager by its ID."""
        if self._lock is None:
            raise SessionError(
                message=(
                    "SessionManager has not been activated. It must be used as an "
                    "asynchronous context manager (`async with ...`)."
                )
            )

        async with self._lock:
            session = self._sessions.pop(session_id, None)
            if session:
                self._stats["total_closed"] += 1
                self._update_stats_unsafe()
                logger.debug("Removed session %s (total: %d)", session_id, len(self._sessions))
            return session

    async def cleanup_closed_sessions(self) -> int:
        """Find and remove any sessions that are marked as closed."""
        if self._lock is None:
            raise SessionError(
                message=(
                    "SessionManager has not been activated. It must be used as an "
                    "asynchronous context manager (`async with ...`)."
                )
            )

        closed_session_ids = []
        async with self._lock:
            for session_id, session in list(self._sessions.items()):
                if session.is_closed:
                    closed_session_ids.append(session_id)
                    del self._sessions[session_id]

            if closed_session_ids:
                self._stats["total_closed"] += len(closed_session_ids)
                self._update_stats_unsafe()
                logger.debug("Cleaned up %d closed sessions.", len(closed_session_ids))

        return len(closed_session_ids)

    async def get_all_sessions(self) -> list[WebTransportSession]:
        """Retrieve a list of all current sessions."""
        if self._lock is None:
            raise SessionError(
                message=(
                    "SessionManager has not been activated. It must be used as an "
                    "asynchronous context manager (`async with ...`)."
                )
            )

        async with self._lock:
            return list(self._sessions.values())

    def get_session_count(self) -> int:
        """Get the current number of active sessions (non-locking)."""
        return len(self._sessions)

    async def get_sessions_by_state(self, *, state: SessionState) -> list[WebTransportSession]:
        """Retrieve sessions that are in a specific state."""
        if self._lock is None:
            raise SessionError(
                message=(
                    "SessionManager has not been activated. It must be used as an "
                    "asynchronous context manager (`async with ...`)."
                )
            )

        async with self._lock:
            return [session for session in self._sessions.values() if session.state == state]

    async def get_stats(self) -> dict[str, Any]:
        """Get detailed statistics about the managed sessions."""
        if self._lock is None:
            raise SessionError(
                message=(
                    "SessionManager has not been activated. It must be used as an "
                    "asynchronous context manager (`async with ...`)."
                )
            )

        async with self._lock:
            states: dict[str, int] = defaultdict(int)
            for session in self._sessions.values():
                states[session.state] += 1
            return {
                **self._stats,
                "active_sessions": len(self._sessions),
                "states": dict(states),
                "max_sessions": self._max_sessions,
            }

    def _on_cleanup_done(self, task: asyncio.Task[None]) -> None:
        """Callback to ensure cleanup is triggered when the cleanup task finishes."""
        if self._is_shutting_down:
            return

        if not task.cancelled():
            if exc := task.exception():
                logger.error("Session cleanup task finished unexpectedly: %s.", exc, exc_info=exc)

        asyncio.create_task(self.shutdown())

    async def _periodic_cleanup(self) -> None:
        """Periodically run the cleanup process to remove closed sessions."""
        try:
            while True:
                try:
                    await self.cleanup_closed_sessions()
                except Exception as e:
                    logger.error("Session cleanup cycle failed: %s", e, exc_info=e)

                await asyncio.sleep(self._cleanup_interval)
        except (Exception, asyncio.CancelledError):
            pass

    def _start_cleanup_task(self) -> None:
        """Start the background task for periodic cleanup if not running."""
        if self._cleanup_task is None or self._cleanup_task.done():
            self._cleanup_task = asyncio.create_task(self._periodic_cleanup())
            self._cleanup_task.add_done_callback(self._on_cleanup_done)

    def _update_stats_unsafe(self) -> None:
        """Update internal statistics (must be called within a lock)."""
        current_count = len(self._sessions)
        self._stats["current_count"] = current_count
        self._stats["max_concurrent"] = max(self._stats["max_concurrent"], current_count)



"""WebTransport Exceptions."""

from __future__ import annotations

from typing import Any

from pywebtransport.constants import ErrorCodes
from pywebtransport.types import SessionState, StreamState

__all__ = [
    "AuthenticationError",
    "CertificateError",
    "ClientError",
    "ConfigurationError",
    "ConnectionError",
    "DatagramError",
    "FlowControlError",
    "HandshakeError",
    "ProtocolError",
    "SerializationError",
    "ServerError",
    "SessionError",
    "StreamError",
    "TimeoutError",
    "WebTransportError",
    "certificate_not_found",
    "connection_timeout",
    "datagram_too_large",
    "get_error_category",
    "invalid_config",
    "is_fatal_error",
    "is_retriable_error",
    "protocol_violation",
    "session_not_ready",
    "stream_closed",
]


class WebTransportError(Exception):
    """The base exception for all WebTransport errors."""

    def __init__(
        self,
        message: str,
        *,
        error_code: int | None = None,
        details: dict[str, Any] | None = None,
    ) -> None:
        """Initialize the WebTransport error."""
        super().__init__(message)
        self.message = message
        self.error_code = error_code or ErrorCodes.INTERNAL_ERROR
        self.details = details or {}

    def to_dict(self) -> dict[str, Any]:
        """Convert the exception to a dictionary for serialization."""
        return {
            "type": self.__class__.__name__,
            "message": self.message,
            "error_code": self.error_code,
            "details": self.details,
        }

    def __str__(self) -> str:
        """Return a simple string representation of the error."""
        return f"[{hex(self.error_code)}] {self.message}"

    def __repr__(self) -> str:
        """Return a detailed string representation of the error."""
        return f"{self.__class__.__name__}(message='{self.message}', error_code={self.error_code})"


class AuthenticationError(WebTransportError):
    """An exception for authentication-related errors."""

    def __init__(
        self,
        message: str,
        *,
        error_code: int | None = None,
        auth_method: str | None = None,
        details: dict[str, Any] | None = None,
    ) -> None:
        """Initialize the authentication error."""
        super().__init__(
            message=message,
            error_code=error_code or ErrorCodes.APP_AUTHENTICATION_FAILED,
            details=details,
        )
        self.auth_method = auth_method

    def to_dict(self) -> dict[str, Any]:
        """Convert the exception to a dictionary."""
        data = super().to_dict()
        data["auth_method"] = self.auth_method
        return data


class CertificateError(WebTransportError):
    """An exception for certificate-related errors."""

    def __init__(
        self,
        message: str,
        *,
        error_code: int | None = None,
        certificate_path: str | None = None,
        certificate_error: str | None = None,
        details: dict[str, Any] | None = None,
    ) -> None:
        """Initialize the certificate error."""
        super().__init__(
            message=message,
            error_code=error_code or ErrorCodes.APP_AUTHENTICATION_FAILED,
            details=details,
        )
        self.certificate_path = certificate_path
        self.certificate_error = certificate_error

    def to_dict(self) -> dict[str, Any]:
        """Convert the exception to a dictionary."""
        data = super().to_dict()
        data["certificate_path"] = self.certificate_path
        data["certificate_error"] = self.certificate_error
        return data


class ClientError(WebTransportError):
    """An exception for client-specific errors."""

    def __init__(
        self,
        message: str,
        *,
        error_code: int | None = None,
        target_url: str | None = None,
        details: dict[str, Any] | None = None,
    ) -> None:
        """Initialize the client error."""
        super().__init__(
            message=message,
            error_code=error_code or ErrorCodes.APP_INVALID_REQUEST,
            details=details,
        )
        self.target_url = target_url

    def to_dict(self) -> dict[str, Any]:
        """Convert the exception to a dictionary."""
        data = super().to_dict()
        data["target_url"] = self.target_url
        return data


class ConfigurationError(WebTransportError):
    """An exception for configuration-related errors."""

    def __init__(
        self,
        message: str,
        *,
        error_code: int | None = None,
        config_key: str | None = None,
        config_value: Any | None = None,
        details: dict[str, Any] | None = None,
    ) -> None:
        """Initialize the configuration error."""
        super().__init__(
            message=message,
            error_code=error_code or ErrorCodes.APP_INVALID_REQUEST,
            details=details,
        )
        self.config_key = config_key
        self.config_value = config_value

    def to_dict(self) -> dict[str, Any]:
        """Convert the exception to a dictionary."""
        data = super().to_dict()
        data["config_key"] = self.config_key
        data["config_value"] = self.config_value
        return data


class ConnectionError(WebTransportError):
    """An exception for connection-related errors."""

    def __init__(
        self,
        message: str,
        *,
        error_code: int | None = None,
        remote_address: tuple[str, int] | None = None,
        details: dict[str, Any] | None = None,
    ) -> None:
        """Initialize the connection error."""
        super().__init__(
            message=message,
            error_code=error_code or ErrorCodes.CONNECTION_REFUSED,
            details=details,
        )
        self.remote_address = remote_address

    def to_dict(self) -> dict[str, Any]:
        """Convert the exception to a dictionary."""
        data = super().to_dict()
        data["remote_address"] = self.remote_address
        return data


class DatagramError(WebTransportError):
    """An exception for datagram-related errors."""

    def __init__(
        self,
        message: str,
        *,
        error_code: int | None = None,
        datagram_size: int | None = None,
        max_size: int | None = None,
        details: dict[str, Any] | None = None,
    ) -> None:
        """Initialize the datagram error."""
        super().__init__(
            message=message,
            error_code=error_code or ErrorCodes.INTERNAL_ERROR,
            details=details,
        )
        self.datagram_size = datagram_size
        self.max_size = max_size

    def to_dict(self) -> dict[str, Any]:
        """Convert the exception to a dictionary."""
        data = super().to_dict()
        data["datagram_size"] = self.datagram_size
        data["max_size"] = self.max_size
        return data


class FlowControlError(WebTransportError):
    """An exception for flow control errors."""

    def __init__(
        self,
        message: str,
        *,
        error_code: int | None = None,
        stream_id: int | None = None,
        limit_exceeded: int | None = None,
        current_value: int | None = None,
        details: dict[str, Any] | None = None,
    ) -> None:
        """Initialize the flow control error."""
        super().__init__(
            message=message,
            error_code=error_code or ErrorCodes.FLOW_CONTROL_ERROR,
            details=details,
        )
        self.stream_id = stream_id
        self.limit_exceeded = limit_exceeded
        self.current_value = current_value

    def to_dict(self) -> dict[str, Any]:
        """Convert the exception to a dictionary."""
        data = super().to_dict()
        data["stream_id"] = self.stream_id
        data["limit_exceeded"] = self.limit_exceeded
        data["current_value"] = self.current_value
        return data


class HandshakeError(WebTransportError):
    """An exception for handshake-related errors."""

    def __init__(
        self,
        message: str,
        *,
        error_code: int | None = None,
        handshake_stage: str | None = None,
        details: dict[str, Any] | None = None,
    ) -> None:
        """Initialize the handshake error."""
        super().__init__(
            message=message,
            error_code=error_code or ErrorCodes.INTERNAL_ERROR,
            details=details,
        )
        self.handshake_stage = handshake_stage

    def to_dict(self) -> dict[str, Any]:
        """Convert the exception to a dictionary."""
        data = super().to_dict()
        data["handshake_stage"] = self.handshake_stage
        return data


class ProtocolError(WebTransportError):
    """An exception for protocol violation errors."""

    def __init__(
        self,
        message: str,
        *,
        error_code: int | None = None,
        frame_type: int | None = None,
        details: dict[str, Any] | None = None,
    ) -> None:
        """Initialize the protocol error."""
        super().__init__(
            message=message,
            error_code=error_code or ErrorCodes.PROTOCOL_VIOLATION,
            details=details,
        )
        self.frame_type = frame_type

    def to_dict(self) -> dict[str, Any]:
        """Convert the exception to a dictionary."""
        data = super().to_dict()
        data["frame_type"] = self.frame_type
        return data


class ServerError(WebTransportError):
    """An exception for server-specific errors."""

    def __init__(
        self,
        message: str,
        *,
        error_code: int | None = None,
        bind_address: tuple[str, int] | None = None,
        details: dict[str, Any] | None = None,
    ) -> None:
        """Initialize the server error."""
        super().__init__(
            message=message,
            error_code=error_code or ErrorCodes.APP_SERVICE_UNAVAILABLE,
            details=details,
        )
        self.bind_address = bind_address

    def to_dict(self) -> dict[str, Any]:
        """Convert the exception to a dictionary."""
        data = super().to_dict()
        data["bind_address"] = self.bind_address
        return data


class SessionError(WebTransportError):
    """An exception for WebTransport session errors."""

    def __init__(
        self,
        message: str,
        *,
        session_id: str | None = None,
        error_code: int | None = None,
        session_state: SessionState | None = None,
        details: dict[str, Any] | None = None,
    ) -> None:
        """Initialize the session error."""
        super().__init__(
            message=message,
            error_code=error_code or ErrorCodes.INTERNAL_ERROR,
            details=details,
        )
        self.session_id = session_id
        self.session_state = session_state

    def to_dict(self) -> dict[str, Any]:
        """Convert the exception to a dictionary."""
        data = super().to_dict()
        data["session_id"] = self.session_id
        data["session_state"] = self.session_state
        return data


class StreamError(WebTransportError):
    """An exception for stream-related errors."""

    def __init__(
        self,
        message: str,
        *,
        stream_id: int | None = None,
        error_code: int | None = None,
        stream_state: StreamState | None = None,
        details: dict[str, Any] | None = None,
    ) -> None:
        """Initialize the stream error."""
        super().__init__(
            message=message,
            error_code=error_code or ErrorCodes.STREAM_STATE_ERROR,
            details=details,
        )
        self.stream_id = stream_id
        self.stream_state = stream_state

    def to_dict(self) -> dict[str, Any]:
        """Convert the exception to a dictionary."""
        data = super().to_dict()
        data["stream_id"] = self.stream_id
        data["stream_state"] = self.stream_state
        return data

    def __str__(self) -> str:
        """Return a simple string representation of the error."""
        base_msg = super().__str__()
        if self.stream_id is not None:
            return f"{base_msg} (stream_id={self.stream_id})"
        return base_msg


class SerializationError(WebTransportError):
    """An exception for serialization or deserialization errors."""

    def __init__(
        self,
        message: str,
        *,
        error_code: int | None = None,
        details: dict[str, Any] | None = None,
        original_exception: Exception | None = None,
    ) -> None:
        """Initialize the serialization error."""
        super().__init__(
            message=message,
            error_code=error_code or ErrorCodes.INTERNAL_ERROR,
            details=details,
        )
        self.original_exception = original_exception

    def to_dict(self) -> dict[str, Any]:
        """Convert the exception to a dictionary."""
        data = super().to_dict()
        data["original_exception"] = str(self.original_exception)
        return data


class TimeoutError(WebTransportError):
    """An exception for timeout-related errors."""

    def __init__(
        self,
        message: str,
        *,
        error_code: int | None = None,
        timeout_duration: float | None = None,
        operation: str | None = None,
        details: dict[str, Any] | None = None,
    ) -> None:
        """Initialize the timeout error."""
        super().__init__(
            message=message,
            error_code=error_code or ErrorCodes.APP_CONNECTION_TIMEOUT,
            details=details,
        )
        self.timeout_duration = timeout_duration
        self.operation = operation

    def to_dict(self) -> dict[str, Any]:
        """Convert the exception to a dictionary."""
        data = super().to_dict()
        data["timeout_duration"] = self.timeout_duration
        data["operation"] = self.operation
        return data


_ERROR_CATEGORY_MAP: dict[type[Exception], str] = {
    AuthenticationError: "authentication",
    CertificateError: "certificate",
    ClientError: "client",
    ConfigurationError: "configuration",
    ConnectionError: "connection",
    DatagramError: "datagram",
    FlowControlError: "flow_control",
    HandshakeError: "handshake",
    ProtocolError: "protocol",
    SerializationError: "serialization",
    ServerError: "server",
    SessionError: "session",
    StreamError: "stream",
    TimeoutError: "timeout",
}


def certificate_not_found(*, path: str) -> CertificateError:
    """Create a certificate not found error."""
    return CertificateError(
        message=f"Certificate file not found: {path}",
        certificate_path=path,
        certificate_error="file_not_found",
    )


def connection_timeout(*, timeout_duration: float, operation: str = "connect") -> TimeoutError:
    """Create a connection timeout error."""
    return TimeoutError(
        message=f"Connection timeout after {timeout_duration}s during {operation}",
        timeout_duration=timeout_duration,
        operation=operation,
    )


def datagram_too_large(*, size: int, max_size: int) -> DatagramError:
    """Create a datagram too large error."""
    return DatagramError(
        message=f"Datagram size {size} exceeds maximum {max_size}",
        datagram_size=size,
        max_size=max_size,
    )


def get_error_category(*, exception: Exception) -> str:
    """Get a simple string category for an exception for logging or monitoring."""
    for exc_type, category in _ERROR_CATEGORY_MAP.items():
        if isinstance(exception, exc_type):
            return category
    return "unknown"


def invalid_config(*, key: str, value: Any, reason: str) -> ConfigurationError:
    """Create an invalid configuration error."""
    return ConfigurationError(
        message=f"Invalid configuration for '{key}': {reason}",
        config_key=key,
        config_value=value,
    )


def is_fatal_error(*, exception: Exception) -> bool:
    """Check if an error is fatal and should terminate the connection."""
    match exception:
        case WebTransportError(error_code=code):
            fatal_codes = {
                ErrorCodes.PROTOCOL_VIOLATION,
                ErrorCodes.FRAME_ENCODING_ERROR,
                ErrorCodes.CRYPTO_BUFFER_EXCEEDED,
                ErrorCodes.APP_AUTHENTICATION_FAILED,
                ErrorCodes.APP_PERMISSION_DENIED,
            }
            return code in fatal_codes
        case _:
            return True


def is_retriable_error(*, exception: Exception) -> bool:
    """Check if an error is transient and the operation can be retried."""
    match exception:
        case WebTransportError(error_code=code):
            retriable_codes = {
                ErrorCodes.APP_CONNECTION_TIMEOUT,
                ErrorCodes.APP_SERVICE_UNAVAILABLE,
                ErrorCodes.FLOW_CONTROL_ERROR,
            }
            return code in retriable_codes
        case _:
            return False


def protocol_violation(*, message: str, frame_type: int | None = None) -> ProtocolError:
    """Create a protocol violation error."""
    return ProtocolError(
        message=message,
        frame_type=frame_type,
        error_code=ErrorCodes.PROTOCOL_VIOLATION,
    )


def session_not_ready(*, session_id: str, current_state: SessionState) -> SessionError:
    """Create a session not ready error."""
    return SessionError(
        message=f"Session {session_id} not ready, current state: {current_state}",
        session_id=session_id,
        session_state=current_state,
    )


def stream_closed(*, stream_id: int, reason: str = "Stream was closed") -> StreamError:
    """Create a stream closed error."""
    return StreamError(
        message=f"Stream {stream_id} closed: {reason}",
        stream_id=stream_id,
        error_code=ErrorCodes.STREAM_STATE_ERROR,
    )



"""WebTransport Application Framework."""

from __future__ import annotations

import asyncio
from collections.abc import Callable
from types import TracebackType
from typing import Any, Self, TypeVar

from pywebtransport.config import ServerConfig
from pywebtransport.connection import WebTransportConnection
from pywebtransport.events import Event
from pywebtransport.server.middleware import MiddlewareManager
from pywebtransport.server.router import RequestRouter
from pywebtransport.server.server import WebTransportServer
from pywebtransport.session import WebTransportSession
from pywebtransport.types import EventType, MiddlewareProtocol, SessionHandler
from pywebtransport.utils import get_logger

__all__ = ["ServerApp"]

logger = get_logger(name="server.app")

F = TypeVar("F", bound=Callable[..., Any])


class ServerApp:
    """A high-level WebTransport application with routing and middleware."""

    def __init__(self, *, config: ServerConfig | None = None) -> None:
        """Initialize the server application."""
        self._server = WebTransportServer(config=config)
        self._router = RequestRouter()
        self._middleware_manager = MiddlewareManager()
        self._stateful_middleware: list[Any] = []
        self._startup_handlers: list[Callable[[], Any]] = []
        self._shutdown_handlers: list[Callable[[], Any]] = []
        self._server.on(event_type=EventType.SESSION_REQUEST, handler=self._handle_session_request)

    @property
    def server(self) -> WebTransportServer:
        """Get the underlying WebTransportServer instance."""
        return self._server

    async def __aenter__(self) -> Self:
        """Enter the async context and run startup procedures."""
        await self._server.__aenter__()
        await self.startup()
        logger.info("ServerApp started.")
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        """Exit the async context and run shutdown procedures."""
        await self.shutdown()
        await self._server.close()
        logger.info("ServerApp stopped.")

    def run(self, *, host: str | None = None, port: int | None = None, **kwargs: Any) -> None:
        """Run the server application in a new asyncio event loop."""
        final_host = host if host is not None else self.server.config.bind_host
        final_port = port if port is not None else self.server.config.bind_port

        async def main() -> None:
            async with self:
                await self.serve(host=final_host, port=final_port, **kwargs)

        try:
            asyncio.run(main())
        except KeyboardInterrupt:
            logger.info("Server stopped by user.")

    async def serve(self, *, host: str | None = None, port: int | None = None, **kwargs: Any) -> None:
        """Start the server and serve forever."""
        final_host = host if host is not None else self.server.config.bind_host
        final_port = port if port is not None else self.server.config.bind_port
        await self._server.listen(host=final_host, port=final_port)
        await self._server.serve_forever()

    async def shutdown(self) -> None:
        """Run all registered shutdown handlers and exit stateful middleware."""
        for handler in self._shutdown_handlers:
            if asyncio.iscoroutinefunction(handler):
                await handler()
            else:
                handler()
        for middleware in reversed(self._stateful_middleware):
            if hasattr(middleware, "__aexit__"):
                await middleware.__aexit__(exc_type=None, exc_val=None, exc_tb=None)

    async def startup(self) -> None:
        """Run all registered startup handlers and enter stateful middleware."""
        for middleware in self._stateful_middleware:
            if hasattr(middleware, "__aenter__"):
                await middleware.__aenter__()
        for handler in self._startup_handlers:
            if asyncio.iscoroutinefunction(handler):
                await handler()
            else:
                handler()

    def add_middleware(self, *, middleware: MiddlewareProtocol) -> None:
        """Add a middleware to the processing chain."""
        self._middleware_manager.add_middleware(middleware=middleware)
        if hasattr(middleware, "__aenter__") and hasattr(middleware, "__aexit__"):
            self._stateful_middleware.append(middleware)

    def middleware(self, middleware_func: MiddlewareProtocol) -> MiddlewareProtocol:
        """Register a middleware function."""
        self.add_middleware(middleware=middleware_func)
        return middleware_func

    def on_shutdown(self, handler: F) -> F:
        """Register a handler to run on application shutdown."""
        self._shutdown_handlers.append(handler)
        return handler

    def on_startup(self, handler: F) -> F:
        """Register a handler to run on application startup."""
        self._startup_handlers.append(handler)
        return handler

    def pattern_route(self, *, pattern: str) -> Callable[[SessionHandler], SessionHandler]:
        """Register a session handler for a URL pattern."""

        def decorator(handler: SessionHandler) -> SessionHandler:
            self._router.add_pattern_route(pattern=pattern, handler=handler)
            return handler

        return decorator

    def route(self, *, path: str) -> Callable[[SessionHandler], SessionHandler]:
        """Register a session handler for a specific path."""

        def decorator(handler: SessionHandler) -> SessionHandler:
            self._router.add_route(path=path, handler=handler)
            return handler

        return decorator

    async def _handle_session_request(self, event: Event) -> None:
        """Handle an incoming session request event from the server."""
        if not isinstance(event.data, dict):
            logger.warning("Session request event data is not a dictionary")
            return

        connection = event.data.get("connection")
        if not isinstance(connection, WebTransportConnection) or not connection.protocol_handler:
            logger.warning("Invalid connection object in session request")
            return

        session_id = event.data.get("session_id")
        stream_id = event.data.get("stream_id")
        if not session_id or stream_id is None:
            if not session_id:
                logger.warning("Missing session_id in session request")
            else:
                logger.warning("Missing stream_id in session request")
            return

        if not connection.is_connected:
            logger.warning("Connection %s is not in connected state", connection.connection_id)
            return

        logger.info(
            "Processing session request: session_id=%s, stream_id=%s",
            session_id,
            stream_id,
        )
        try:
            session_info = connection.protocol_handler.get_session_info(session_id=session_id)
            if not session_info:
                logger.error("Session info not found for session %s", session_id)
                return

            config = connection.config
            if not isinstance(config, ServerConfig):
                logger.error(
                    "Connection %s has a non-server config, which is unexpected.",
                    connection.connection_id,
                )
                connection.protocol_handler.close_webtransport_session(
                    session_id=session_id,
                    code=1,
                    reason="Internal server configuration error",
                )
                return

            session = WebTransportSession(
                connection=connection,
                session_id=session_id,
                max_streams=config.max_streams_per_connection,
                max_incoming_streams=config.max_incoming_streams,
                stream_cleanup_interval=config.stream_cleanup_interval,
            )
            await session.initialize()
            session._path = session_info.path
            session._headers = session_info.headers
            session._control_stream_id = stream_id

            if self.server.session_manager:
                await self.server.session_manager.add_session(session=session)

            logger.info("Processing session request for path '%s'", session.path)
            if not await self._middleware_manager.process_request(session=session):
                logger.warning(
                    "Session request for path '%s' rejected by middleware.",
                    session.path,
                )
                connection.protocol_handler.close_webtransport_session(
                    session_id=session_id, code=403, reason="Rejected by middleware"
                )
                return

            handler = self._router.route_request(session=session)
            if not handler:
                logger.warning("No route found for path: %s", session.path)
                connection.protocol_handler.close_webtransport_session(
                    session_id=session_id, code=404, reason="Route not found"
                )
                return

            logger.info(
                "Routing session request for path '%s' to handler '%s'",
                session.path,
                handler.__name__,
            )
            connection.protocol_handler.accept_webtransport_session(stream_id=stream_id, session_id=session_id)
            logger.info("Handler task created for session %s", session_id)

            async def run_handler(*, h: SessionHandler, s: WebTransportSession) -> None:
                try:
                    logger.debug("Handler starting for session %s", s.session_id)
                    await h(s)
                    logger.debug("Handler completed for session %s", s.session_id)
                except Exception as handler_error:
                    logger.error(
                        "Handler error for session %s: %s",
                        s.session_id,
                        handler_error,
                        exc_info=True,
                    )
                finally:
                    if not s.is_closed:
                        logger.debug("Closing session %s", s.session_id)
                        await s.close()

            asyncio.create_task(run_handler(h=handler, s=session))

        except Exception as e:
            logger.error(
                "Error handling session request for session %s: %s",
                session_id,
                e,
                exc_info=True,
            )
            try:
                if connection.protocol_handler:
                    connection.protocol_handler.close_webtransport_session(
                        session_id=session_id,
                        code=1,
                        reason="Internal server error",
                    )
            except Exception:
                pass



"""WebTransport stream pooling for efficient stream reuse."""

from __future__ import annotations

import asyncio
from types import TracebackType
from typing import TYPE_CHECKING, Self

from pywebtransport.exceptions import StreamError
from pywebtransport.stream.stream import WebTransportStream
from pywebtransport.utils import get_logger

if TYPE_CHECKING:
    from pywebtransport.session import WebTransportSession


__all__ = ["StreamPool"]

logger = get_logger(name="stream.pool")


class StreamPool:
    """Manages a pool of reusable WebTransport streams for a session."""

    def __init__(
        self,
        session: WebTransportSession,
        *,
        pool_size: int = 10,
        maintenance_interval: float = 60.0,
    ) -> None:
        """Initialize the stream pool."""
        if pool_size <= 0:
            raise ValueError("Pool size must be a positive integer.")

        self._session = session
        self._pool_size = pool_size
        self._maintenance_interval = maintenance_interval
        self._available: list[WebTransportStream] = []
        self._total_managed_streams = 0
        self._condition: asyncio.Condition | None = None
        self._maintenance_task: asyncio.Task[None] | None = None

    @classmethod
    def create(
        cls,
        *,
        session: WebTransportSession,
        pool_size: int = 10,
    ) -> Self:
        """Create a new stream pool instance."""
        return cls(session=session, pool_size=pool_size)

    async def __aenter__(self) -> Self:
        """Enter the async context and initialize the pool."""
        self._condition = asyncio.Condition()
        await self._initialize_pool()
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        """Exit the async context and close all streams in the pool."""
        await self.close_all()

    async def close_all(self) -> None:
        """Close all idle streams and shut down the pool."""
        if self._condition is None:
            raise StreamError(
                message=(
                    "StreamPool has not been activated. It must be used as an "
                    "asynchronous context manager (`async with ...`)."
                )
            )

        if self._maintenance_task and not self._maintenance_task.done():
            self._maintenance_task.cancel()
            try:
                await self._maintenance_task
            except asyncio.CancelledError:
                pass

        streams_to_close: list[WebTransportStream] = []
        async with self._condition:
            streams_to_close.extend(self._available)
            self._available.clear()
            self._total_managed_streams = 0

        if streams_to_close:
            try:
                async with asyncio.TaskGroup() as tg:
                    for s in streams_to_close:
                        tg.create_task(s.close())
            except* Exception as eg:
                logger.error(
                    "Errors occurred while closing pooled streams: %s",
                    eg.exceptions,
                    exc_info=eg,
                )
            logger.info("Closed %d idle streams from the pool.", len(streams_to_close))

    async def get_stream(self, *, timeout: float | None = None) -> WebTransportStream:
        """Get a stream from the pool, creating a new one if necessary."""
        if self._condition is None:
            raise StreamError(
                message=(
                    "StreamPool has not been activated. It must be used as an "
                    "asynchronous context manager (`async with ...`)."
                )
            )

        async with self._condition:
            while True:
                while self._available:
                    stream = self._available.pop(0)
                    if not stream.is_closed:
                        logger.debug("Reusing stream %d from pool.", stream.stream_id)
                        return stream
                    else:
                        logger.debug("Discarding stale stream %d from pool.", stream.stream_id)
                        self._total_managed_streams -= 1

                if self._total_managed_streams < self._pool_size:
                    self._total_managed_streams += 1
                    break

                logger.debug("Pool is full. Waiting for a stream to be returned.")
                try:
                    await asyncio.wait_for(self._condition.wait(), timeout=timeout)
                except asyncio.TimeoutError:
                    raise StreamError(message=f"Timeout waiting for a stream from the pool after {timeout}s.") from None

        try:
            logger.debug("Creating new stream as pool was empty or depleted.")
            return await self._session.create_bidirectional_stream()
        except Exception:
            async with self._condition:
                self._total_managed_streams -= 1
                self._condition.notify()
            raise

    async def return_stream(self, *, stream: WebTransportStream) -> None:
        """Return a stream to the pool for potential reuse."""
        if self._condition is None:
            raise StreamError(
                message=(
                    "StreamPool has not been activated. It must be used as an "
                    "asynchronous context manager (`async with ...`)."
                )
            )

        should_close = stream.is_closed
        async with self._condition:
            if not should_close:
                if len(self._available) >= self._pool_size:
                    should_close = True
                else:
                    self._available.append(stream)
                    logger.debug("Returned stream %d to pool.", stream.stream_id)
                    self._condition.notify()

            if should_close:
                if self._total_managed_streams > len(self._available):
                    self._total_managed_streams -= 1
                    self._condition.notify()

        if should_close:
            await stream.close()

    async def _fill_pool(self) -> None:
        """Create new streams concurrently until the pool reaches its target size."""
        if self._condition is None:
            return
        if self._session.is_closed:
            logger.warning("Session closed, cannot replenish stream pool.")
            return

        created_streams: list[WebTransportStream] = []
        needed = 0
        async with self._condition:
            needed = self._pool_size - self._total_managed_streams
            if needed <= 0:
                return
            self._total_managed_streams += needed

        try:
            async with asyncio.TaskGroup() as tg:
                tasks = [tg.create_task(self._session.create_bidirectional_stream()) for _ in range(needed)]
            created_streams = [task.result() for task in tasks if task.done() and not task.exception()]
        except* Exception as eg:
            logger.error(
                "Failed to create %d streams for the pool: %s",
                needed,
                eg.exceptions,
                exc_info=eg,
            )
            async with self._condition:
                self._total_managed_streams -= needed
                self._condition.notify_all()

        if created_streams:
            async with self._condition:
                successful_count = len(created_streams)
                if successful_count < needed:
                    self._total_managed_streams -= needed - successful_count
                self._available.extend(created_streams)
                if successful_count > 0:
                    self._condition.notify_all()

    async def _initialize_pool(self) -> None:
        """Initialize the pool by pre-filling it with new streams."""
        if self._condition is None:
            return

        needs_fill = False
        async with self._condition:
            if self._total_managed_streams == 0:
                needs_fill = True

        if needs_fill:
            try:
                await self._fill_pool()
                self._start_maintenance_task()
                logger.info(
                    "Stream pool initialized with %d streams.",
                    self._total_managed_streams,
                )
            except Exception as e:
                logger.error("Error initializing stream pool: %s", e, exc_info=True)
                await self.close_all()

    async def _maintain_pool_loop(self) -> None:
        """Periodically check and replenish the stream pool."""
        if self._condition is None:
            return

        try:
            while True:
                await asyncio.sleep(self._maintenance_interval)
                needs_fill = False
                async with self._condition:
                    if self._total_managed_streams < self._pool_size:
                        logger.debug(
                            "Replenishing pool. Size (%d) is below target (%d).",
                            self._total_managed_streams,
                            self._pool_size,
                        )
                        needs_fill = True
                if needs_fill:
                    await self._fill_pool()
        except asyncio.CancelledError:
            logger.info("Stream pool maintenance task cancelled.")
        except Exception as e:
            logger.error("Stream pool maintenance task crashed: %s", e, exc_info=e)

    def _start_maintenance_task(self) -> None:
        """Start the periodic pool maintenance task if not already running."""
        if self._maintenance_task is None:
            coro = self._maintain_pool_loop()
            try:
                self._maintenance_task = asyncio.create_task(coro)
            except RuntimeError:
                coro.close()
                self._maintenance_task = None
                logger.warning("Could not start pool maintenance task: no running event loop.")
